"""
This type stub file was generated by pyright.
"""

from collections import OrderedDict
from collections.abc import Iterator, Mapping
from typing import Optional

__all__ = ["Pins", "PinsN", "DiffPairs", "DiffPairsN", "Attrs", "Clock", "Subsignal", "Resource", "Connector"]
class Pins:
    def __init__(self, names: str, *, dir: str=..., invert: bool=..., conn: Optional[tuple[str, int|str]]=..., assert_width: Optional[int]=...) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def __iter__(self) -> Iterator[str]:
        ...
    
    def map_names(self, mapping: Mapping[str, str], resource) -> list[str]:
        ...
    
    def __repr__(self) -> str:
        ...
    


def PinsN(*args, **kwargs) -> Pins:
    ...

class DiffPairs:
    def __init__(self, p: str, n: str, *, dir: str=..., invert: bool=..., conn: Optional[tuple[str, int|str]]=..., assert_width: Optional[int]=...) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def __iter__(self) -> Iterator[tuple[str, str]]:
        ...
    
    def __repr__(self) -> str:
        ...
    


def DiffPairsN(*args, **kwargs) -> DiffPairs:
    ...

class Attrs(OrderedDict):
    def __init__(self, **attrs) -> None:
        ...
    
    def __repr__(self) -> str:
        ...
    


class Clock:
    def __init__(self, frequency: float | int) -> None:
        ...
    
    @property
    def period(self) -> float:
        ...
    
    def __repr__(self) -> str:
        ...
    


class Subsignal:
    def __init__(self, name: str, *args: Pins | DiffPairs | Subsignal | Attrs | Clock) -> None:
        ...
    
    def __repr__(self) -> str:
        ...
    


class Resource(Subsignal):
    @classmethod
    def family(cls, name_or_number: str | int, number: Optional[int]=..., *, ios, default_name: str, name_suffix: str=...): # -> Self@Resource:
        ...
    
    def __init__(self, name, number, *args) -> None:
        ...
    
    def __repr__(self) -> str:
        ...
    


class Connector:
    def __init__(self, name: str, number: int, io: str | dict[str, str], *, conn: Optional[tuple[str, int|str]]=...) -> None:
        ...
    
    def __repr__(self) -> str:
        ...
    
    def __len__(self) -> int:
        ...
    
    def __iter__(self) -> Iterator[tuple[str, str]]:
        ...
    





"""
This type stub file was generated by pyright.
"""

__all__ = ["BaseProcess", "BaseSignalState", "BaseMemoryState", "BaseEngineState", "BaseEngine"]

class BaseProcess:
    __slots__ = ...

    runnable: bool
    critical: bool
    
    def reset(self): # -> None:
        ...
    
    def run(self):
        ...


class BaseSignalState:
    __slots__ = ...
    signal = ...
    curr = ...
    next = ...
    def update(self, value):
        ...
    

class BaseMemoryState:
    __slots__ = ...
    memory = ...

    def read(self, addr):
        ...

    def write(self, addr, value, mask=None):
        ...


class BaseEngineState:
    def reset(self):
        ...
    
    def get_signal(self, signal):
        ...
    
    def get_memory(self, signal):
        ...
    
    slots = ...

    def set_delay_waker(self, interval, waker):
        ...

    def add_signal_waker(self, signal, waker):
        ...

    def add_memory_waker(self, memory, waker):
        ...


class BaseEngine:
    @property
    def state(self) -> BaseEngineState:
        ...

    @property
    def now(self):
        ...

    def reset(self):
        ...

    def add_clock_process(self, clock, *, phase, period):
        ...

    def add_async_process(self, simulator, process):
        ...

    def add_async_testbench(self, simulator, process, *, background):
        ...

    def add_trigger_combination(self, combination, *, oneshot):
        ...
    
    def advance(self):
        ...
    
    def write_vcd(self, *, vcd_file, gtkw_file, traces):
        ...
    





"""
This type stub file was generated by pyright.
"""

from contextlib import contextmanager
from abc import ABCMeta, abstractmethod

__all__ = ["BuildPlan", "BuildProducts", "LocalBuildProducts", "RemoteSSHBuildProducts"]
class BuildPlan:
    def __init__(self, script) -> None:
        """A build plan.

        Parameters
        ----------
        script : str
            The base name (without extension) of the script that will be executed.
        """
        ...
    
    def add_file(self, filename, content): # -> None:
        """
        Add ``content``, which can be a :class:`str`` or :class:`bytes`, to the build plan
        as ``filename``. The file name can be a relative path with directories separated by
        forward slashes (``/``).
        """
        ...
    
    def digest(self, size=...): # -> bytes:
        """
        Compute a `digest`, a short byte sequence deterministically and uniquely identifying
        this build plan.
        """
        ...
    
    def archive(self, file): # -> None:
        """
        Archive files from the build plan into ``file``, which can be either a filename, or
        a file-like object. The produced archive is deterministic: exact same files will
        always produce exact same archive.
        """
        ...
    
    def execute_local(self, root=..., *, run_script=...): # -> LocalBuildProducts:
        """
        Execute build plan using the local strategy. Files from the build plan are placed in
        the build root directory ``root``, and, if ``run_script`` is ``True``, the script
        appropriate for the platform (``{script}.bat`` on Windows, ``{script}.sh`` elsewhere) is
        executed in the build root.

        Returns :class:`LocalBuildProducts`.
        """
        ...
    
    def execute_remote_ssh(self, *, connect_to=..., root, run_script=...): # -> RemoteSSHBuildProducts:
        """
        Execute build plan using the remote SSH strategy. Files from the build
        plan are transferred via SFTP to the directory ``root`` on a  remote
        server. If ``run_script`` is ``True``, the ``paramiko`` SSH client will
        then run ``{script}.sh``. ``root`` can either be an absolute or
        relative (to the login directory) path.

        ``connect_to`` is a dictionary that holds all input arguments to
        ``paramiko``'s ``SSHClient.connect``
        (`documentation <http://docs.paramiko.org/en/stable/api/client.html#paramiko.client.SSHClient.connect>`_).
        At a minimum, the ``hostname`` input argument must be supplied in this
        dictionary as the remote server.

        Returns :class:`RemoteSSHBuildProducts`.
        """
        ...
    
    def execute(self): # -> LocalBuildProducts:
        """
        Execute build plan using the default strategy. Use one of the ``execute_*`` methods
        explicitly to have more control over the strategy.
        """
        ...
    


class BuildProducts(metaclass=ABCMeta):
    @abstractmethod
    def get(self, filename, mode=...): # -> None:
        """
        Extract ``filename`` from build products, and return it as a :class:`bytes` (if ``mode``
        is ``"b"``) or a :class:`str` (if ``mode`` is ``"t"``).
        """
        ...
    
    @contextmanager
    def extract(self, *filenames): # -> Generator[Unknown | list[Unknown] | None, Unknown, Unknown]:
        """
        Extract ``filenames`` from build products, place them in an OS-specific temporary file
        location, with the extension preserved, and delete them afterwards. This method is used
        as a context manager, e.g.: ::

            with products.extract("bitstream.bin", "programmer.cfg") \
                    as bitstream_filename, config_filename:
                subprocess.check_call(["program", "-c", config_filename, bitstream_filename])
        """
        ...
    


class LocalBuildProducts(BuildProducts):
    def __init__(self, root) -> None:
        ...
    
    def get(self, filename, mode=...): # -> Any:
        ...
    


class RemoteSSHBuildProducts(BuildProducts):
    def __init__(self, connect_to, root) -> None:
        ...
    
    def get(self, filename, mode=...): # -> str | bytes:
        ...
    





"""
This type stub file was generated by pyright.
"""

from ..hdl import *
from ..build import *

class LatticeECP5Platform(TemplatedPlatform):
    """
    .. rubric:: Trellis toolchain

    Required tools:
        * ``yosys``
        * ``nextpnr-ecp5``
        * ``ecppack``

    The environment is populated by running the script specified in the environment variable
    ``AMARANTH_ENV_TRELLIS``, if present.

    Available overrides:
        * ``verbose``: enables logging of informational messages to standard error.
        * ``read_verilog_opts``: adds options for ``read_verilog`` Yosys command.
        * ``synth_opts``: adds options for ``synth_ecp5`` Yosys command.
        * ``script_after_read``: inserts commands after ``read_ilang`` in Yosys script.
        * ``script_after_synth``: inserts commands after ``synth_ecp5`` in Yosys script.
        * ``yosys_opts``: adds extra options for ``yosys``.
        * ``nextpnr_opts``: adds extra options for ``nextpnr-ecp5``.
        * ``ecppack_opts``: adds extra options for ``ecppack``.
        * ``add_preferences``: inserts commands at the end of the LPF file.

    Build products:
        * ``{{name}}.rpt``: Yosys log.
        * ``{{name}}.json``: synthesized RTL.
        * ``{{name}}.tim``: nextpnr log.
        * ``{{name}}.config``: ASCII bitstream.
        * ``{{name}}.bit``: binary bitstream.
        * ``{{name}}.svf``: JTAG programming vector.

    .. rubric:: Diamond toolchain

    Required tools:
        * ``pnmainc``
        * ``ddtcmd``

    The environment is populated by running the script specified in the environment variable
    ``AMARANTH_ENV_DIAMOND``, if present. On Linux, diamond_env as provided by Diamond
    itself is a good candidate. On Windows, the following script (named ``diamond_env.bat``,
    for instance) is known to work::

        @echo off
        set PATH=C:\\lscc\\diamond\\%DIAMOND_VERSION%\\bin\\nt64;%PATH%

    Available overrides:
        * ``script_project``: inserts commands before ``prj_project save`` in Tcl script.
        * ``script_after_export``: inserts commands after ``prj_run Export`` in Tcl script.
        * ``add_preferences``: inserts commands at the end of the LPF file.
        * ``add_constraints``: inserts commands at the end of the XDC file.

    Build products:
        * ``{{name}}_impl/{{name}}_impl.htm``: consolidated log.
        * ``{{name}}.bit``: binary bitstream.
        * ``{{name}}.svf``: JTAG programming vector.
    """
    toolchain = ...
    device = ...
    package = ...
    speed = ...
    grade = ...
    _nextpnr_device_options = ...
    _nextpnr_package_options = ...
    _trellis_required_tools = ...
    _trellis_file_templates = ...
    _trellis_command_templates = ...
    _diamond_required_tools = ...
    _diamond_file_templates = ...
    _diamond_command_templates = ...
    def __init__(self, *, toolchain=...) -> None:
        ...
    
    @property
    def required_tools(self): # -> list[str]:
        ...
    
    @property
    def file_templates(self): # -> dict[str, str]:
        ...
    
    @property
    def command_templates(self): # -> list[str]:
        ...
    
    @property
    def default_clk_constraint(self): # -> Clock:
        ...
    
    def create_missing_domain(self, name): # -> Module | None:
        ...
    
    _single_ended_io_types = ...
    _differential_io_types = ...
    def should_skip_port_component(self, port, attrs, component): # -> bool:
        ...
    
    def get_input(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_output(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_tristate(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_input_output(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_diff_input(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_diff_output(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_diff_tristate(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_diff_input_output(self, pin, port, attrs, invert): # -> Module:
        ...
    





"""
This type stub file was generated by pyright.
"""

from abc import ABCMeta, abstractmethod
from ._ast import *
from ._cd import *
from ._ir import *

"""
This type stub file was generated by pyright.
"""
__all__ = ["ValueVisitor", "ValueTransformer", "StatementVisitor", "StatementTransformer", "FragmentTransformer", "TransformedElaboratable", "DomainCollector", "DomainRenamer", "DomainLowerer", "ResetInserter", "EnableInserter"]
class ValueVisitor(metaclass=ABCMeta):
    @abstractmethod
    def on_Const(self, value):
        ...
    
    @abstractmethod
    def on_Signal(self, value):
        ...
    
    @abstractmethod
    def on_ClockSignal(self, value):
        ...
    
    @abstractmethod
    def on_ResetSignal(self, value):
        ...
    
    @abstractmethod
    def on_AnyValue(self, value):
        ...
    
    @abstractmethod
    def on_Operator(self, value):
        ...
    
    @abstractmethod
    def on_Slice(self, value):
        ...
    
    @abstractmethod
    def on_Part(self, value):
        ...
    
    @abstractmethod
    def on_Concat(self, value):
        ...
    
    @abstractmethod
    def on_SwitchValue(self, value):
        ...
    
    @abstractmethod
    def on_Initial(self, value):
        ...
    
    def on_unknown_value(self, value):
        ...
    
    def replace_value_src_loc(self, value, new_value):
        ...
    
    def on_value(self, value):
        ...
    
    def __call__(self, value):
        ...
    


class ValueTransformer(ValueVisitor):
    def on_Const(self, value):
        ...
    
    def on_Signal(self, value):
        ...
    
    def on_ClockSignal(self, value):
        ...
    
    def on_ResetSignal(self, value):
        ...
    
    def on_AnyValue(self, value):
        ...
    
    def on_Operator(self, value):
        ...
    
    def on_Slice(self, value):
        ...
    
    def on_Part(self, value):
        ...
    
    def on_Concat(self, value):
        ...
    
    def on_SwitchValue(self, value):
        ...
    
    def on_Initial(self, value):
        ...
    


class StatementVisitor(metaclass=ABCMeta):
    @abstractmethod
    def on_Assign(self, stmt):
        ...
    
    @abstractmethod
    def on_Print(self, stmt):
        ...
    
    @abstractmethod
    def on_Property(self, stmt):
        ...
    
    @abstractmethod
    def on_Switch(self, stmt):
        ...
    
    @abstractmethod
    def on_statements(self, stmts):
        ...
    
    def on_unknown_statement(self, stmt):
        ...
    
    def replace_statement_src_loc(self, stmt, new_stmt):
        ...
    
    def on_statement(self, stmt):
        ...
    
    def __call__(self, stmt):
        ...
    


class StatementTransformer(StatementVisitor):
    def on_value(self, value):
        ...
    
    def on_Format(self, format):
        ...
    
    def on_Assign(self, stmt):
        ...
    
    def on_Print(self, stmt):
        ...
    
    def on_Property(self, stmt):
        ...
    
    def on_Switch(self, stmt):
        ...
    
    def on_statements(self, stmts):
        ...
    


class FragmentTransformer:
    def map_subfragments(self, fragment, new_fragment):
        ...
    
    def map_ports(self, fragment, new_fragment):
        ...
    
    def map_domains(self, fragment, new_fragment):
        ...
    
    def map_statements(self, fragment, new_fragment):
        ...
    
    def map_memory_ports(self, fragment, new_fragment):
        ...
    
    def on_fragment(self, fragment):
        ...
    
    def __call__(self, value, *, src_loc_at=...):
        ...
    


class TransformedElaboratable(Elaboratable):
    def __init__(self, elaboratable, *, src_loc_at=...) -> None:
        ...
    
    def __getattr__(self, attr):
        ...
    
    def elaborate(self, platform):
        ...
    


class DomainCollector(ValueVisitor, StatementVisitor):
    def __init__(self) -> None:
        ...
    
    def on_ignore(self, value):
        ...
    
    on_Const = ...
    on_Signal = ...
    on_AnyValue = ...
    def on_ClockSignal(self, value):
        ...
    
    def on_ResetSignal(self, value):
        ...
    
    def on_Operator(self, value):
        ...
    
    def on_Slice(self, value):
        ...
    
    def on_Part(self, value):
        ...
    
    def on_Concat(self, value):
        ...
    
    def on_SwitchValue(self, value):
        ...
    
    def on_Initial(self, value):
        ...
    
    def on_Format(self, format):
        ...
    
    def on_Assign(self, stmt):
        ...
    
    def on_Print(self, stmt):
        ...
    
    def on_Property(self, stmt):
        ...
    
    def on_Switch(self, stmt):
        ...
    
    def on_statements(self, stmts):
        ...
    
    def on_fragment(self, fragment):
        ...
    
    def __call__(self, fragment):
        ...
    


class DomainRenamer(FragmentTransformer, ValueTransformer, StatementTransformer):
    def __init__(self, domain_map) -> None:
        ...
    
    def on_ClockSignal(self, value):
        ...
    
    def on_ResetSignal(self, value):
        ...
    
    def map_domains(self, fragment, new_fragment):
        ...
    
    def map_statements(self, fragment, new_fragment):
        ...
    
    def map_memory_ports(self, fragment, new_fragment):
        ...
    


class DomainLowerer(FragmentTransformer, ValueTransformer, StatementTransformer):
    def __init__(self, domains=...) -> None:
        ...
    
    def replace_value_src_loc(self, value, new_value):
        ...
    
    def on_ClockSignal(self, value):
        ...
    
    def on_ResetSignal(self, value):
        ...
    
    def on_fragment(self, fragment):
        ...
    


class _ControlInserter(FragmentTransformer):
    def __init__(self, controls) -> None:
        ...
    
    def on_fragment(self, fragment):
        ...
    
    def __call__(self, value, *, src_loc_at=...):
        ...
    


class ResetInserter(_ControlInserter):
    ...


class EnableInserter(_ControlInserter):
    def on_fragment(self, fragment):
        ...
    





"""
This type stub file was generated by pyright.
"""

from typing import Any
from contextlib import contextmanager
from ..hdl import *
from ._base import *

__all__ = ["PySimEngine"]

class _VCDWriter:
    @staticmethod
    def decode_to_vcd(signal, value):
        ...
    
    def __init__(self, state, design, *, vcd_file, gtkw_file=..., traces=..., fs_per_delta=..., processes=...) -> None:
        ...
    
    def update(self, timestamp, signal, value): # -> None:
        ...
    
    def close(self, timestamp):
        ...
    
    vcd_file: Any
    gtkw_file: Any
    gtkw_save: Any
    gtkw_signal_names: Any
    gtkw_memory_names: Any
    vcd_writer: Any

class _PyTimeline:
    def __init__(self) -> None:
        ...
    
    def reset(self): # -> None:
        ...
    
    def set_waker(self, interval, waker):
        ...

    def advance(self): # -> bool:
        ...
    


class _PySignalState(BaseSignalState):
    __slots__ = ...
    def __init__(self, signal, pending) -> None:
        ...

    def reset(self):
        ...

    def add_waker(self, waker):
        ...

    def update(self, value):
        ...
    
    def commit(self): # -> bool:
        ...
    


class _PyEngineState(BaseEngineState):
    def __init__(self) -> None:
        ...
    
    def reset(self): # -> None:
        ...
    
    def get_signal(self, signal): # -> int:
        ...
    
    def get_memory(self, memory): # -> int:
        ...
    
    def set_delay_waker(self, interval, waker):
        ...

    def add_signal_waker(self, signal, waker):
        ...

    def add_memory_waker(self, memory, waker):
        ...

    def commit(self, changed=...): # -> bool:
        ...
    


class PySimEngine(BaseEngine):
    def __init__(self, fragment) -> None:
        ...
    
    @property
    def state(self) -> BaseEngineState:
        ...

    @property
    def now(self): # -> int | None:
        ...
    
    def reset(self): # -> None:
        ...
    
    def advance(self): # -> bool:
        ...

    def add_clock_process(self, clock, *, phase, period):
        ...

    def add_async_process(self, simulator, process):
        ...

    def add_async_testbench(self, simulator, process, *, background):
        ...

    def add_trigger_combination(self, combination, *, oneshot):
        ...

    def get_value(self, expr):
        ...

    def set_value(self, expr, value):
        ...

    def step_design(self):
        ...

    @contextmanager
    def write_vcd(self, *, vcd_file, gtkw_file, traces, fs_per_delta): # -> Generator[None, None, None]:
        ...
    





"""
This type stub file was generated by pyright.
"""

from contextlib import _GeneratorContextManager, contextmanager
from typing import Callable, ContextManager, Iterator, NoReturn, OrderedDict, ParamSpec, TypeVar, Optional, Self
from amaranth_types import HasElaborate
from ._ast import *
from ._ast import Flattenable
from ._ir import *
from ._cd import *
from ._xfrm import *

__all__ = ["SyntaxError", "SyntaxWarning", "Module"]


_P = ParamSpec("_P")
_T = TypeVar("_T")


class SyntaxError(Exception):
    ...


class SyntaxWarning(Warning):
    ...


class _ModuleBuilderProxy:
    def __init__(self, builder, depth) -> None:
        ...
    


class _ModuleBuilderDomain(_ModuleBuilderProxy):
    def __init__(self, builder, depth, domain) -> None:
        ...
    
    def __iadd__(self, assigns: StatementLike) -> Self:
        ...
    


class _ModuleBuilderDomains(_ModuleBuilderProxy):
    def __getattr__(self, name: str) -> _ModuleBuilderDomain:
        ...
    
    def __getitem__(self, name: str) -> _ModuleBuilderDomain:
        ...
    
    def __setattr__(self, name: str, value: _ModuleBuilderDomain) -> None:
        ...
    
    def __setitem__(self, name: str, value: _ModuleBuilderDomain) -> None:
        ...
    


class _ModuleBuilderRoot:
    d: _ModuleBuilderDomains

    def __init__(self, builder, depth) -> None:
        ...
    
    def __getattr__(self, name) -> NoReturn:
        ...
    


class _ModuleBuilderSubmodules:
    def __init__(self, builder) -> None:
        ...
    
    def __iadd__(self, modules: Flattenable[HasElaborate]) -> Self:
        ...
    
    def __setattr__(self, name: str, submodule: HasElaborate) -> None:
        ...
    
    def __setitem__(self, name: str, value: HasElaborate) -> None:
        ...
    
    def __getattr__(self, name: str):
        ...
    
    def __getitem__(self, name: str):
        ...
    


class _ModuleBuilderDomainSet:
    def __init__(self, builder) -> None:
        ...
    
    def __iadd__(self, domains: Flattenable[ClockDomain]) -> Self:
        ...
    
    def __setattr__(self, name: str, domain: ClockDomain) -> None:
        ...
    


class _GuardedContextManager(_GeneratorContextManager[_T]):
    def __init__(self, keyword, func, args, kwds) -> None:
        ...
    
    def __bool__(self) -> NoReturn:
        ...
    

def _guardedcontextmanager(keyword: str) -> Callable[[Callable[_P, Iterator[_T]]], Callable[_P, _GuardedContextManager[_T]]]:
    ...


class FSM:
    def __init__(self, state: Signal, encoding: OrderedDict[str, int], decoding: OrderedDict[int, str]) -> None:
        ...
    
    def ongoing(self, name: str) -> Value:
        ...
    


class Module(_ModuleBuilderRoot, Elaboratable):
    submodules: _ModuleBuilderSubmodules
    domains: _ModuleBuilderDomainSet
    d: _ModuleBuilderDomains
    domain: _ModuleBuilderDomains

    @classmethod
    def __init_subclass__(cls):
        ...
    
    def __init__(self) -> None:
        ...
    
    @_guardedcontextmanager("If")
    def If(self, cond: ValueLike) -> Iterator[None]:
        ...
    
    @_guardedcontextmanager("Elif")
    def Elif(self, cond: ValueLike) -> Iterator[None]:
        ...
    
    @_guardedcontextmanager("Else")
    def Else(self) -> Iterator[None]:
        ...
    
    @contextmanager
    def Switch(self, test: ValueLike) -> Iterator[None]:
        ...
    
    @contextmanager
    def Case(self, *patterns: SwitchKey) -> Iterator[None]:
        ...
    
    def Default(self) -> ContextManager[None]:
        ...
    
    @contextmanager
    def FSM(self, init: Optional[str] = ..., domain: str = ..., name: str = ...) -> Iterator[FSM]:
        ...
    
    @contextmanager
    def State(self, name: str) -> Iterator[None]:
        ...
    
    @property
    def next(self) -> NoReturn:
        ...
    
    @next.setter
    def next(self, name: str) -> None:
        ...
    
    def elaborate(self, platform) -> Fragment:
        ...
    





from collections.abc import AsyncGenerator
from contextlib import contextmanager
from typing import Any, Never, Optional, overload
from ..hdl import *
from ._base import BaseProcess, BaseEngine
from amaranth_types import ValueLike


__all__ = [
    "DomainReset", "BrokenTrigger",
    "SampleTrigger", "ChangedTrigger", "EdgeTrigger", "DelayTrigger",
    "TriggerCombination", "TickTrigger",
    "SimulatorContext", "ProcessContext", "TestbenchContext", "AsyncProcess",
]


class DomainReset(Exception):
    """Exception raised when the domain of a a tick trigger that is repeatedly awaited has its
    reset asserted."""


class BrokenTrigger(Exception):
    """Exception raised when a trigger that is repeatedly awaited in an `async for` loop has
    a matching event occur while the body of the `async for` loop is executing."""


class SampleTrigger:
    def __init__(self, value: ValueLike):
        ...


class ChangedTrigger:
    def __init__(self, signal: ValueLike):
        ...

    @property
    def value(self) -> Signal:
        ...


class EdgeTrigger:
    def __init__(self, signal: ValueLike, polarity: int):
        ...


class DelayTrigger:
    def __init__(self, interval: int | float):
        ...


class TriggerCombination:
    def __init__(self, engine: BaseEngine, process: BaseProcess, *,
            triggers: tuple[DelayTrigger|ChangedTrigger|SampleTrigger|EdgeTrigger, ...] = ...):
        ...

    def sample(self, *values: ValueLike) -> TriggerCombination:
        ...

    def changed(self, *signals: ValueLike) -> TriggerCombination:
        ...

    def edge(self, signal: ValueLike, polarity: int) -> TriggerCombination:
        ...

    def posedge(self, signal: ValueLike) -> TriggerCombination:
        ...

    def negedge(self, signal: ValueLike) -> TriggerCombination:
        ...

    def delay(self, interval: int | float) -> TriggerCombination:
        ...

    def __await__(self):
        ...

    async def __aiter__(self) -> AsyncGenerator:
        ...


class TickTrigger:
    def __init__(self, engine: BaseEngine, process: BaseProcess, *,
            domain: ClockDomain, sampled: tuple[ValueLike] = ...):
        ...

    def sample(self, *values: ValueLike) -> TickTrigger:
        ...

    async def until(self, condition: ValueLike):
        ...

    async def repeat(self, count: int):
        ...

    def __await__(self):
        ...

    async def __aiter__(self) -> AsyncGenerator:
        ...


class SimulatorContext:
    def __init__(self, design, engine: BaseEngine, process: BaseProcess):
        ...

    def delay(self, interval: int | float) -> TriggerCombination:
        ...

    def changed(self, *signals: ValueLike) -> TriggerCombination:
        ...

    def edge(self, signal: ValueLike, polarity: int) -> TriggerCombination:
        ...

    def posedge(self, signal: ValueLike) -> TriggerCombination:
        ...

    @overload
    def tick(self, domain: str = ..., *, context: Optional[Elaboratable] = ...) -> TickTrigger: ...

    @overload
    def tick(self, domain: ClockDomain) -> TickTrigger: ...

    def tick(self, domain: str | ClockDomain = ..., *, context: Optional[Elaboratable] = ...) -> TickTrigger:
        ...

    @contextmanager
    def critical(self):
        ...
    
    @overload
    def get(self, expr: Value) -> int: ...

    @overload
    def get(self, expr: ValueCastable) -> Any: ...

    def get(self, expr: Value | ValueCastable) -> Any:
        ...

    @overload
    def set(self, expr: Value, value: int) -> None: ...

    @overload
    def set(self, expr: ValueCastable, value: Any) -> None: ...

    def set(self, expr: Value | ValueCastable, value: Any) -> None:
        ...


class ProcessContext(SimulatorContext):
    def get(self, expr: ValueLike) -> Never:
        ...

    @overload
    def set(self, expr: Value, value: int) -> None: ...

    @overload
    def set(self, expr: ValueCastable, value: Any) -> None: ...

    def set(self, expr: Value | ValueCastable, value: Any) -> None:
        ...


class TestbenchContext(SimulatorContext):
    @overload
    def get(self, expr: Value) -> int: ...

    @overload
    def get(self, expr: ValueCastable) -> Any: ...

    def get(self, expr: Value | ValueCastable) -> Any:
        ...

    @overload
    def set(self, expr: Value, value: int) -> None: ...

    @overload
    def set(self, expr: ValueCastable, value: Any) -> None: ...

    def set(self, expr: Value | ValueCastable, value: Any) -> None:
        ...


class AsyncProcess(BaseProcess):
    def __init__(self, design, engine: BaseEngine, constructor, *, testbench: bool, background):
        ...

    def reset(self) -> None:
        ...

    def run(self) -> None:
        ...



"""
This type stub file was generated by pyright.
"""

from abc import ABCMeta, abstractmethod
from collections.abc import Callable, MutableMapping, MutableSequence, MutableSet
from typing import Any, Generic, Iterable, Iterator, Mapping, NoReturn, Optional, Sequence, TypeVar, final, overload
from enum import Enum, EnumMeta
from amaranth_types import ValueLike, FlatValueLike, ShapeLike, FlatShapeLike, StatementLike
from amaranth.lib.data import View
from amaranth_types.types import IOValueLike


__all__: list[str]  = [
    "SyntaxError", "SyntaxWarning",
    "Shape", "signed", "unsigned", "ShapeCastable", "ShapeLike",
    "Value", "Const", "C", "AnyValue", "AnyConst", "AnySeq", "Operator", "Mux", "Part", "Slice", "Cat", "Concat", "SwitchValue",
    "Array", "ArrayProxy",
    "Signal", "ClockSignal", "ResetSignal",
    "ValueCastable", "ValueLike",
    "Initial",
    "Format",
    "Statement", "Switch",
    "Property", "Assign", "Print", "Assert", "Assume", "Cover",
    "IOValue", "IOPort", "IOConcat", "IOSlice",
    "SignalKey", "SignalDict", "SignalSet",
]


T = TypeVar("T")
U = TypeVar("U")
_T_ShapeCastable = TypeVar("_T_ShapeCastable", bound=ShapeCastable, covariant=True)
Flattenable = T | Iterable[Flattenable[T]]
SwitchKey = str | int | Enum


class SyntaxError(Exception):
    pass
  

class SyntaxWarning(Warning):
    pass


class DUID:
    """Deterministic Unique IDentifier."""
    __next_uid = ...
    def __init__(self) -> None:
        ...
    

class ShapeCastable(Generic[U]):
    def __init__(self, *args, **kwargs) -> None:
        ...

    def __init_subclass__(cls, **kwargs) -> None:
        ...

    def as_shape(self, *args, **kwargs) -> Shape:
        ...

    def __call__(self, target: ValueLike) -> U:
        ...

    def const(self, *args, **kwargs) -> Const:
        ...

    def from_bits(self, raw: int):
        ...

    def format(self, obj: ValueLike, spec: str) -> Format:
        ...


class Shape:
    """Bit width and signedness of a va"""
    @property
    def width(self) -> int:
        ...

    @property
    def signed(self) -> int:
        ...

    def __init__(self, width: int =..., signed: bool =...) -> None:
        ...
    
    def __iter__(self) -> Iterator[int | bool]:
        ...
    
    @staticmethod
    def cast(obj: ShapeLike, *, src_loc_at=...) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    
    def __eq__(self, other) -> bool:
        ...

    def __hash__(self) -> int:
        ...
    


def unsigned(width) -> Shape:
    """Shorthand for ``Shape(width, sig"""
    ...

def signed(width) -> Shape:
    """Shorthand for ``Shape(width, sig"""
    ...

class Value(metaclass=ABCMeta):
    @staticmethod
    def cast(obj: ValueLike) -> Value:
        """Converts ``obj`` to an Amaranth """
        ...
    
    def __init__(self, *, src_loc_at=...) -> None:
        ...
    
    def __bool__(self) -> NoReturn:
        ...
    
    def __invert__(self) -> Value:
        ...
    
    def __pos__(self) -> Value:
        ...
    
    def __neg__(self) -> Value:
        ...
    
    def __add__(self, other: ValueLike) -> Value:
        ...
    
    def __radd__(self, other: ValueLike) -> Value:
        ...
    
    def __sub__(self, other: ValueLike) -> Value:
        ...
    
    def __rsub__(self, other: ValueLike) -> Value:
        ...
    
    def __mul__(self, other: ValueLike) -> Value:
        ...
    
    def __rmul__(self, other: ValueLike) -> Value:
        ...
    
    def __mod__(self, other: ValueLike) -> Value:
        ...
    
    def __rmod__(self, other: ValueLike) -> Value:
        ...
    
    def __floordiv__(self, other: ValueLike) -> Value:
        ...
    
    def __rfloordiv__(self, other: ValueLike) -> Value:
        ...
    
    def __lshift__(self, other: ValueLike) -> Value:
        ...
    
    def __rlshift__(self, other: ValueLike) -> Value:
        ...
    
    def __rshift__(self, other: ValueLike) -> Value:
        ...
    
    def __rrshift__(self, other: ValueLike) -> Value:
        ...
    
    def __and__(self, other: ValueLike) -> Value:
        ...
    
    def __rand__(self, other: ValueLike) -> Value:
        ...
    
    def __xor__(self, other: ValueLike) -> Value:
        ...
    
    def __rxor__(self, other: ValueLike) -> Value:
        ...
    
    def __or__(self, other: ValueLike) -> Value:
        ...
    
    def __ror__(self, other: ValueLike) -> Value:
        ...
    
    def __eq__(self, other: ValueLike) -> Value:
        ...
    
    def __ne__(self, other: ValueLike) -> Value:
        ...
    
    def __lt__(self, other: ValueLike) -> Value:
        ...
    
    def __le__(self, other: ValueLike) -> Value:
        ...
    
    def __gt__(self, other: ValueLike) -> Value:
        ...
    
    def __ge__(self, other: ValueLike) -> Value:
        ...
    
    def __abs__(self) -> Value:
        ...
    
    def __len__(self) -> int:
        ...
    
    def __getitem__(self, key: int | slice) -> Value:
        ...

    def __contains__(self, other) -> NoReturn:
        ...

    def __format__(self, format_desc) -> NoReturn:
        ...

    def as_unsigned(self) -> Value:
        """Conversion to unsigned.

       """
        ...
    
    def as_signed(self) -> Value:
        """Conversion to signed.

        R"""
        ...
    
    def bool(self) -> Value:
        """Conversion to boolean.

        """
        ...
    
    def any(self) -> Value:
        """Check if any bits are ``1``.

  """
        ...
    
    def all(self) -> Value:
        """Check if all bits are ``1``.

  """
        ...
    
    def xor(self) -> Value:
        """Compute pairwise exclusive-or of"""
        ...
    
    def bit_select(self, offset: ValueLike, width: int) -> Value:
        """Part-select with bit granularity"""
        ...
    
    def word_select(self, offset: ValueLike, width: int) -> Value:
        """Part-select with word granularit"""
        ...
    
    def matches(self, *patterns) -> Value:
        """Pattern matching.

        Match"""
        ...
    
    def shift_left(self, amount: int) -> Value:
        """Shift left by constant amount.

"""
        ...
    
    def shift_right(self, amount: int) -> Value:
        """Shift right by constant amount.
"""
        ...
    
    def rotate_left(self, amount: int) -> Value:
        """Rotate left by constant amount.
"""
        ...
    
    def rotate_right(self, amount: int) -> Value:
        """Rotate right by constant amount."""
        ...
    
    def replicate(self, count : int) -> Value:
        """Replication.

        A ``Value`` is replicated (repeated) several times to be used
        on the RHS of assignments::

            len(v.replicate(n)) == len(v) * n

        Parameters
        ----------
        count : int
            Number of replications.

        Returns
        -------
        Value, out
            Replicated value.
        """
        ...

    def eq(self, value: ValueLike) -> Assign:
        """Assignment.

        Parameters
"""
        ...
    
    @abstractmethod
    def shape(self) -> Shape:
        """Bit width and signedness of a va"""
        ...
    
    __hash__ = ...


class _ConstMeta(ABCMeta):
    def __call__(cls, value: int | ShapeCastable, shape: Optional[ShapeLike] = ..., src_loc_at=0, **kwargs):
        ...


@final
class Const(Value, metaclass=_ConstMeta):
    """A constant, literal integer valu"""
    src_loc = ...
    @staticmethod
    def cast(obj: ValueLike) -> Const:
        ...

    def __init__(self, value: int, shape: Optional[ShapeLike] =..., *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...

    @property
    def value(self) -> int:
        ...

    def __repr__(self) -> str:
        ...
    


C = Const
class AnyValue(Value, DUID):
    def __init__(self, shape: ShapeLike, *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    


@final
class AnyConst(AnyValue):
    def __repr__(self) -> str:
        ...
    


@final
class AnySeq(AnyValue):
    def __repr__(self) -> str:
        ...
    


@final
class Operator(Value):
    def __init__(self, operator: str, operands: Iterable[ValueLike], *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    


def Mux(sel: ValueLike, val1: ValueLike, val0: ValueLike) -> Value:
    """Choose between two values.

    """
    ...

@final
class Slice(Value):
    def __init__(self, value: ValueLike, start: int, stop: int, *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...

    value: Value
    start: int
    stop: int


@final
class Part(Value):
    def __init__(self, value: Value, offset: ValueLike, width: int, stride: int = ..., *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    


@final
class Cat(Value):
    """Concatenate values.

    Form a """
    def __init__(self, *args: Flattenable[ValueLike], src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...


@final
class Concat(Value):
    def __init__(self, args: Iterable[ValueLike], src_loc_at=0) -> None:
        ...

    @property
    def parts(self) -> tuple[Value, ...]:
        ...

    def shape(self) -> Shape:
        ...

    def __repr__(self) -> str:
        ...


@final
class SwitchValue(Value):
    def __init__(self, test, cases, *, src_loc=None, src_loc_at=0) -> None:
        ...

    @property
    def test(self) -> Value:
        ...

    @property
    def cases(self) -> tuple[tuple[Optional[tuple[str, ...]], Value], ...]:
        ...

    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...


@final
class Repl(Value):
    """Replicate a value

    An input """
    def __init__(self, value: ValueLike, count: int, *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    

class _SignalMeta(ABCMeta):
    @overload
    def __call__(cls, shape: ShapeCastable[T], src_loc_at: int = ..., **kwargs) -> T:
        ...
    
    @overload
    def __call__(cls, shape: FlatShapeLike = ..., src_loc_at: int = ..., **kwargs) -> Signal:
        ...
    
    def __call__(cls, shape: ShapeLike = ..., src_loc_at: int = ..., **kwargs):
        ...


class Signal(Value, DUID, metaclass=_SignalMeta):
    """A varying integer value.

    Pa"""
    def __init__(self, shape: Optional[ShapeLike] = ..., *, name: Optional[str] = ..., init: int | Enum = ..., reset_less: bool = ..., attrs: dict = ..., decoder: type[Enum] | Callable[[int], str] = ..., src_loc_at=...) -> None:
        ...

    @overload
    @staticmethod
    def like(other: View[_T_ShapeCastable], *, name: Optional[str] = ..., name_suffix: Optional[str] =..., src_loc_at=..., **kwargs) -> View[_T_ShapeCastable]:
        ...

    @overload
    @staticmethod
    def like(other: FlatValueLike, *, name: Optional[str] = ..., name_suffix: Optional[str] =..., src_loc_at=..., **kwargs) -> Signal:
        ...

    @staticmethod
    def like(other: ValueLike, *, name: Optional[str] = ..., name_suffix: Optional[str] =..., src_loc_at=..., **kwargs):
        """Create Signal based on another.
"""
        ...

    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    
    name: str
    decoder: Any


@final
class ClockSignal(Value):
    """Clock signal for a clock domain."""
    def __init__(self, domain: str = ..., *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    


@final
class ResetSignal(Value):
    """Reset signal for a clock domain."""
    def __init__(self, domain: str = ..., allow_reset_less: bool = ..., *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    


class Array(MutableSequence[T]):
    """Addressable multiplexer.

    An"""

    def __init__(self, iterable: Iterable[T] = ()) -> None:
        ...
    
    @overload
    def __getitem__(self, index: int) -> T:
        ...
    
    @overload
    def __getitem__(self, index: Value) -> ArrayProxy:
        ...

    def __getitem__(self, index: int | Value) -> T | ArrayProxy:
        ...
    
    def __len__(self) -> int:
        ...
    
    def __setitem__(self, index: int, value: T) -> None:
        ...
    
    def __delitem__(self, index: int) -> None:
        ...

    def insert(self, index: int, value: T) -> None:
        ...
    
    def __repr__(self) -> str:
        ...
    


@final
class ArrayProxy(Value):
    elems: Sequence[ValueLike]
    index: Value

    def __init__(self, elems: Sequence[ValueLike], index: ValueLike, *, src_loc_at=...) -> None:
        ...
    
    def __getattr__(self, attr: str) -> ArrayProxy:
        ...
    
    def __getitem__(self, index: Value | int | str) -> ArrayProxy:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    


class ValueCastable:
    """Base class for classes which can"""
    def __new__(cls, *args, **kwargs): # -> Self@ValueCastable:
        ...
    
    @staticmethod
    def lowermethod(func): # -> (self: Unknown, *args: Unknown, **kwargs: Unknown) -> Unknown:
        """Decorator to memoize lowering me"""
        ...

    @abstractmethod
    def as_value(self) -> Value:
        ...

    @abstractmethod
    def shape(self) -> ShapeLike:
        ...


@final
class Sample(Value):
    """Value from the past.

    A ``Sa"""
    def __init__(self, expr: ValueLike, clocks: int, domain: Optional[str], *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...
    


def Past(expr: ValueLike, clocks: int = ..., domain: Optional[str] = ...) -> Sample:
    ...

def Stable(expr: ValueLike, clocks: int = ..., domain: Optional[str] = ...) -> Value:
    ...

def Rose(expr: ValueLike, clocks: int = ..., domain: Optional[str] = ...) -> Value:
    ...

def Fell(expr: ValueLike, clocks: int = ..., domain: Optional[str] = ...) -> Value:
    ...

@final
class Initial(Value):
    """Start indicator, for model check"""
    def __init__(self, *, src_loc_at=...) -> None:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def __repr__(self) -> str:
        ...


class _FormatLike:
    def __add__(self, other) -> Format:
        ...

    def __format__(self, format_desc) -> NoReturn:
        ...


@final
class Format(_FormatLike):
    def __init__(self, format: str, *args, **kwargs) -> None:
        ...

    def __repr__(self) -> str:
        ...

    class Enum(_FormatLike):
        def __init__(self, value: ValueLike, /, variants: EnumMeta | Mapping[int, str] | Iterable[tuple[int, str]], *, name: Optional[str] = ...):
            ...

        def __repr__(self) -> str:
            ...
    
    class Struct(_FormatLike):
        def __init__(self, value: ValueLike, /, fields: Mapping[str, _FormatLike] | Iterable[tuple[str, _FormatLike]]):
            ...

        def __repr__(self) -> str:
            ...

    
    class Array(_FormatLike):
        def __init__(self, value: ValueLike, /, fields: Iterable[_FormatLike]):
            ...

        def __repr__(self) -> str:
            ...



class _StatementList(list[Statement]):
    def __repr__(self) -> str:
        ...
    


class Statement:
    def __init__(self, *, src_loc_at=...) -> None:
        ...
    
    @staticmethod
    def cast(obj: StatementLike) -> _StatementList:
        ...
    


@final
class Assign(Statement):
    lhs: Value
    rhs: Value

    def __init__(self, lhs: ValueLike, rhs: ValueLike, *, src_loc_at=...) -> None:
        ...
    
    def __repr__(self) -> str:
        ...
    

@final
class Print(Statement):
    def __init__(self, *args, sep: str = ..., end: str = ..., src_loc_at=...) -> None:
        ...

    @property
    def message(self) -> Format:
        ...
    
    def __repr__(self) -> str:
        ...


class Property(Statement):
    class Kind(Enum):
        Assert = ...
        Assume = ...
        Cover = ...

    def __init__(self, kind: str | Kind, test: ValueLike, message: Optional[str | Format] = ..., *, src_loc_at=...) -> None:
        ...
    
    def __repr__(self) -> str:
        ...


def Assert(test: ValueLike, message: Optional[str | Format], *, src_loc_at=...) -> Property:
    ...


def Assume(test: ValueLike, message: Optional[str | Format], *, src_loc_at=...) -> Property:
    ...


def Cover(test: ValueLike, message: Optional[str | Format], *, src_loc_at=...) -> Property:
    ...


class Switch(Statement):
    def __init__(self, test: ValueLike, cases: Mapping[SwitchKey, StatementLike], *, src_loc=..., src_loc_at=..., case_src_locs=...) -> None:
        ...
    
    def __repr__(self) -> str:
        ...
    

class IOValue(metaclass=ABCMeta):
    @staticmethod
    def cast(obj: IOValueLike) -> IOValue:
        ...

    def __init__(self, *, src_loc_at : int = ...) -> None:
        ...

    @property
    @abstractmethod
    def metadata(self) -> tuple:
        ...

    def __getitem__(self, key: int | slice) -> IOValue:
        ...


@final
class IOPort(IOValue):
    def __init__(self, width: int, *, name: Optional[str] = ..., attrs: Optional[dict] = ..., metadata: Optional[tuple] = ..., src_loc_at: int = ...) -> None:
        ...

    def __len__(self) -> int:
        ...

    @property
    def width(self) -> int:
        ...

    @property
    def attrs(self) -> dict:
        ...

    @property
    def metadata(self) -> tuple:
        ...

    def __repr__(self) -> str:
        ...


@final
class IOConcat(IOValue):
    def __init__(self, parts: Iterable[IOValue], src_loc_at: int = ...) -> None:
        ...

    @property
    def parts(self) -> tuple[IOValue]:
        ...

    def __len__(self) -> int:
        ...

    @property
    def metadata(self) -> tuple:
        ...

    def __repr__(self) -> str:
        ...


@final
class IOSlice(IOValue):
    def __init__(self, value: IOValueLike, start: int, stop: int, *, src_loc_at: int = ...) -> None:
        ...

    @property
    def value(self) -> IOValue:
        ...

    @property
    def start(self) -> int:
        ...

    @property
    def stop(self) -> int:
        ...

    def __len__(self) -> int:
        ...

    @property
    def metadata(self) -> tuple:
        ...

    def __repr__(self) -> str:
        ...


class _MappedKeyCollection(metaclass=ABCMeta):
    ...


class _MappedKeyDict(MutableMapping, _MappedKeyCollection):
    def __init__(self, pairs=...) -> None:
        ...
    
    def __getitem__(self, key):
        ...
    
    def __setitem__(self, key, value): # -> None:
        ...
    
    def __delitem__(self, key): # -> None:
        ...
    
    def __iter__(self): # -> Generator[None, None, None]:
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def __len__(self): # -> int:
        ...
    
    def __repr__(self): # -> str:
        ...
    


class _MappedKeySet(MutableSet, _MappedKeyCollection):
    def __init__(self, elements=...) -> None:
        ...
    
    def add(self, value): # -> None:
        ...
    
    def update(self, values): # -> None:
        ...
    
    def discard(self, value): # -> None:
        ...
    
    def __contains__(self, value): # -> bool:
        ...
    
    def __iter__(self): # -> Generator[None, None, None]:
        ...
    
    def __len__(self): # -> int:
        ...
    
    def __repr__(self): # -> str:
        ...
    


class SignalKey:
    def __init__(self, signal: Signal | ClockSignal | ResetSignal) -> None:
        ...
    
    def __hash__(self) -> int:
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def __lt__(self, other: SignalKey) -> bool:
        ...
    
    def __repr__(self) -> str:
        ...
    


class SignalDict(_MappedKeyDict):
    _map_key = ...
    _unmap_key = ...


class SignalSet(_MappedKeySet):
    _map_key = ...
    _unmap_key = ...





"""
This type stub file was generated by pyright.
"""

from .core import *

__all__ = [
    "DomainReset", "BrokenTrigger", "Simulator",
    # deprecated
    "Settle", "Delay", "Tick", "Passive", "Active",
]



"""
This type stub file was generated by pyright.
"""

from .hdl import *

__all__ = ["Shape", "unsigned", "signed", "Value", "Const", "C", "Mux", "Cat", "Array", "Signal", "ClockSignal", "ResetSignal", "Module", "ClockDomain", "Elaboratable", "Fragment", "Instance", "Memory", "Record", "DomainRenamer", "ResetInserter", "EnableInserter"]



"""
This type stub file was generated by pyright.
"""

from typing import Any
from ..hdl._ast import *
from ..hdl.rec import *
from ..lib.io import *
from .dsl import *
from collections.abc import Iterable, Iterator

__all__ = ["ResourceError", "ResourceManager"]
class ResourceError(Exception):
    ...


class PortGroup:
    ...


class ResourceManager:
    def __init__(self, resources: Iterable[Resource], connectors: Iterable[Connector]) -> None:
        ...
    
    def add_resources(self, resources: Iterable[Resource]) -> None:
        ...
    
    def add_connectors(self, connectors: Iterable[Connector]) -> None:
        ...
    
    def lookup(self, name: str, number: int=...) -> Resource:
        ...
    
    def request(self, name: str, number: int=..., *, dir=..., xdr=...) -> Pin | PortLike | PortGroup:
        ...
    
    def iter_single_ended_pins(self): # -> Generator[tuple[Unknown, Unknown, Unknown, Unknown], Any, None]:
        ...
    
    def iter_differential_pins(self): # -> Generator[tuple[Unknown, Unknown, Unknown, Unknown], Any, None]:
        ...
    
    def should_skip_port_component(self, port, attrs, component) -> bool:
        ...
    
    def add_clock_constraint(self, clock: Signal, frequency: int | float) -> None:
        ...
    
    def iter_clock_constraints(self) -> Iterator[tuple[Signal, Any, int | float]]:
        ...
    





from typing import (
    Callable,
    Concatenate,
    Generic,
    NoReturn,
    Optional,
    Protocol,
    ParamSpec,
    TypeAlias,
    TypeVar,
    cast,
    Union,
    Any,
    TYPE_CHECKING,
    runtime_checkable,
)
from collections.abc import Coroutine, Generator, Iterable, Iterator
from contextlib import AbstractContextManager
from enum import Enum
from amaranth import *
from amaranth.lib.wiring import Flow, Member
from typing import Sequence
from amaranth.hdl import IOValue, ShapeCastable, ValueCastable
from amaranth.lib.data import Layout
from amaranth.sim._pycoro import Command

if TYPE_CHECKING:
    from amaranth.hdl._ast import Statement
    from amaranth.hdl._dsl import _ModuleBuilderSubmodules, _ModuleBuilderDomainSet, _ModuleBuilderDomain
    import amaranth.hdl._dsl

__all__ = [
    "FragmentLike",
    "ValueLike",
    "FlatValueLike",
    "ShapeLike",
    "FlatShapeLike",
    "StatementLike",
    "LayoutLike",
    "SwitchKey",
    "SrcLoc",
    "TestGenerator",
    "TestCoroutine",
    "_ModuleBuilderDomainsLike",
    "ModuleLike",
    "HasElaborate",
    "AbstractInterface",
    "AbstractSignature",
    "AbstractComponent"
]

T = TypeVar("T")
U = TypeVar("U")
P = ParamSpec("P")

# Types representing Amaranth concepts
FragmentLike: TypeAlias = Fragment | Elaboratable
FlatValueLike: TypeAlias = Value | int | Enum
ValueLike: TypeAlias = FlatValueLike | ValueCastable
IOValueLike: TypeAlias = IOValue | Value
FlatShapeLike: TypeAlias = Shape | int | range | type[Enum]
ShapeLike: TypeAlias = FlatShapeLike | ShapeCastable
StatementLike: TypeAlias = Union["Statement", Iterable["StatementLike"]]
SwitchKey: TypeAlias = str | int | Enum
SrcLoc: TypeAlias = tuple[str, int]
TestGenerator: TypeAlias = Generator[Union[Command, Value, "Statement", None], Any, T]
TestCoroutine: TypeAlias = Coroutine[Any, Any, T]

# Deprecated
LayoutLike: TypeAlias = Layout | Sequence[tuple[str, "ShapeLike | LayoutLike"]]


# Protocols for Amaranth classes
class _ModuleBuilderDomainsLike(Protocol):
    def __getattr__(self, name: str) -> "_ModuleBuilderDomain": ...

    def __getitem__(self, name: str) -> "_ModuleBuilderDomain": ...

    def __setattr__(self, name: str, value: "_ModuleBuilderDomain") -> None: ...

    def __setitem__(self, name: str, value: "_ModuleBuilderDomain") -> None: ...


_T_ModuleBuilderDomains = TypeVar("_T_ModuleBuilderDomains", bound=_ModuleBuilderDomainsLike)


class ModuleLike(Protocol, Generic[_T_ModuleBuilderDomains]):
    submodules: "_ModuleBuilderSubmodules"
    domains: "_ModuleBuilderDomainSet"
    d: _T_ModuleBuilderDomains

    def If(self, cond: ValueLike) -> AbstractContextManager[None]:  # noqa: N802
        ...

    def Elif(self, cond: ValueLike) -> AbstractContextManager[None]:  # noqa: N802
        ...

    def Else(self) -> AbstractContextManager[None]:  # noqa: N802
        ...

    def Switch(self, test: ValueLike) -> AbstractContextManager[None]:  # noqa: N802
        ...

    def Case(self, *patterns: SwitchKey) -> AbstractContextManager[None]:  # noqa: N802
        ...

    def Default(self) -> AbstractContextManager[None]:  # noqa: N802
        ...

    def FSM(  # noqa: N802
        self, init: Optional[str] = ..., domain: str = ..., name: str = ...
    ) -> AbstractContextManager["amaranth.hdl._dsl.FSM"]: ...

    def State(self, name: str) -> AbstractContextManager[None]:  # noqa: N802
        ...

    @property
    def next(self) -> NoReturn: ...

    @next.setter
    def next(self, name: str) -> None: ...


class AbstractSignatureMembers(Protocol):
    def flip(self) -> "AbstractSignatureMembers": ...

    def __eq__(self, other) -> bool: ...

    def __contains__(self, name: str) -> bool: ...

    def __getitem__(self, name: str) -> Member: ...

    def __setitem__(self, name: str, member: Member) -> NoReturn: ...

    def __delitem__(self, name: str) -> NoReturn: ...

    def __iter__(self) -> Iterator[str]: ...

    def __len__(self) -> int: ...

    def flatten(self, *, path: tuple[str | int, ...] = ...) -> Iterator[tuple[tuple[str | int, ...], Member]]: ...

    def create(self, *, path: tuple[str | int, ...] = ..., src_loc_at: int = ...) -> dict[str, Any]: ...

    def __repr__(self) -> str: ...


class AbstractSignature(Protocol):
    def flip(self) -> "AbstractSignature": ...

    @property
    def members(self) -> AbstractSignatureMembers: ...

    def __eq__(self, other) -> bool: ...

    def flatten(self, obj) -> Iterator[tuple[tuple[str | int, ...], Flow, ValueLike]]: ...

    def is_compliant(self, obj, *, reasons: Optional[list[str]] = ..., path: tuple[str, ...] = ...) -> bool: ...

    def create(
        self, *, path: tuple[str | int, ...] = ..., src_loc_at: int = ...
    ) -> "AbstractInterface": ...

    def __repr__(self) -> str: ...


_T_AbstractSignature = TypeVar("_T_AbstractSignature", bound=AbstractSignature, covariant=True)


@runtime_checkable
class AbstractInterface(Protocol, Generic[_T_AbstractSignature]):
    @property
    def signature(self) -> _T_AbstractSignature: ...


class HasElaborate(Protocol):
    def elaborate(self, platform) -> "HasElaborate": ...


class AbstractComponent(AbstractInterface[_T_AbstractSignature], HasElaborate, Protocol):
    pass


def type_self_kwargs_as(as_func: Callable[Concatenate[Any, P], Any]):
    """
    Decorator used to annotate `**kwargs` type to be the same as named arguments from `as_func` method.

    Works only with methods with (self, **kwargs) signature. `self` parameter is also required in `as_func`.
    """

    def return_func(func: Callable[Concatenate[Any, ...], T]) -> Callable[Concatenate[Any, P], T]:
        return cast(Callable[Concatenate[Any, P], T], func)

    return return_func



"""
This type stub file was generated by pyright.
"""

__all__ = ["ClockDomain", "DomainError"]
class DomainError(Exception):
    ...


class ClockDomain:
    """Synchronous domain.

    Paramet"""
    def __init__(self, name=..., *, clk_edge=..., reset_less=..., async_reset=..., local=...) -> None:
        ...
    
    def rename(self, new_name): # -> None:
        ...
    





"""
This type stub file was generated by pyright.
"""

from .. import *
from ..asserts import *
from amaranth_types import HasElaborate
from . import stream, wiring

"""First-in first-out queues."""
__all__ = ["FIFOInterface", "SyncFIFO", "SyncFIFOBuffered", "AsyncFIFO", "AsyncFIFOBuffered"]
class FIFOInterface:
    width: int
    depth: int
    w_data: Signal
    w_rdy: Signal
    w_en: Signal
    w_level: Signal
    r_data: Signal
    r_rdy: Signal
    r_en: Signal
    r_level: Signal
    def __init__(self, *, width: int, depth: int) -> None:
        ...
    
    @property
    def w_stream(self) -> wiring.FlippedSignature[stream.Signature]:
        ...

    @property
    def r_stream(self) -> stream.Signature:
        ...


class SyncFIFO(Elaboratable, FIFOInterface):
    def __init__(self, *, width: int, depth: int) -> None:
        ...
    
    def elaborate(self, platform) -> HasElaborate:
        ...
    


class SyncFIFOBuffered(Elaboratable, FIFOInterface):
    def __init__(self, *, width: int, depth: int) -> None:
        ...
    
    def elaborate(self, platform) -> HasElaborate:
        ...
    


class AsyncFIFO(Elaboratable, FIFOInterface):
    def __init__(self, *, width: int, depth: int, r_domain: str = ..., w_domain: str = ..., exact_depth: bool = ...) -> None:
        ...
    
    def elaborate(self, platform) -> HasElaborate:
        ...
    


class AsyncFIFOBuffered(Elaboratable, FIFOInterface):
    def __init__(self, *, width: int, depth: int, r_domain: str = ..., w_domain: str = ..., exact_depth: bool = ...) -> None:
        ...
    
    def elaborate(self, platform) -> HasElaborate:
        ...
    





"""
This type stub file was generated by pyright.
"""

from ._lattice_ecp5 import LatticeECP5Platform
from ._lattice_ice40 import LatticeICE40Platform

__all__ = ["LatticeECP5Platform", "LatticeICE40Platform"]



import enum
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from typing import Optional

from amaranth_types.types import IOValueLike

from ..hdl import *
from ..lib import wiring


__all__ = [
    "Direction", "PortLike", "SingleEndedPort", "DifferentialPort", "SimulationPort",
    "Buffer", "FFBuffer", "DDRBuffer",
    "Pin",
]


class Direction(enum.Enum):
    """Represents a direction of an I/O port, or of an I/O buffer."""

    #: Input direction (from world to Amaranth design)
    Input  = "i"
    #: Output direction (from Amaranth design to world)
    Output = "o"
    #: Bidirectional (can be switched between input and output)
    Bidir  = "io"

    def __or__(self, other: Direction) -> Direction:
        ...

    def __and__(self, other: Direction) -> Direction:
        ...


class PortLike(metaclass=ABCMeta):
    """Represents an abstract port that can be passed to a buffer.

    The port types supported by most platforms are :class:`SingleEndedPort` and
    :class:`DifferentialPort`. Platforms may define additional custom port types as appropriate.
    """

    @property
    @abstractmethod
    def direction(self) -> Direction:
        """The direction of this port, as :class:`Direction`."""
        ...

    @abstractmethod
    def __len__(self) -> int:
        """Returns the width of this port in bits."""
        ...

    @abstractmethod
    def __getitem__(self, index: int | slice) -> PortLike:
        """Slices the port, returning another :class:`PortLike` with a subset
        of its bits.

        The index can be a :class:`slice` or :class:`int`. If the index is
        an :class:`int`, the result is a single-bit :class:`PortLike`."""
        ...

    @abstractmethod
    def __invert__(self) -> PortLike:
        """Returns a new :class:`PortLike` object like this one, but with inverted polarity.

        The result should be such that using :class:`Buffer` on it is equivalent to using
        :class:`Buffer` on the original, with added inverters on the :py:`i` and :py:`o` ports."""
        ...

    @abstractmethod
    def __add__(self, other) -> PortLike:
        """Concatenates two library I/O ports of the same type.
        The direction of the resulting port is:
        * The same as the direction of both, if the two ports have the same direction.
        * :attr:`Direction.Input` if a bidirectional port is concatenated with an input port.
        * :attr:`Direction.Output` if a bidirectional port is concatenated with an output port.
        Returns
        -------
        :py:`type(self)`
            A new :py:`type(self)` which contains wires from :py:`self` followed by wires
            from :py:`other`, preserving their polarity inversion.
        Raises
        ------
        :exc:`ValueError`
            If an input port is concatenated with an output port.
        :exc:`TypeError`
            If :py:`self` and :py:`other` have different types.
        """
        ...


class SingleEndedPort(PortLike):
    """Represents a single-ended I/O port with optional inversion.

    Parameters
    ----------
    io : :class:`IOValue`
        The raw I/O value being wrapped.
    invert : :class:`bool` or iterable of :class:`bool`
        If true, the electrical state of the physical pin will be opposite from the Amaranth value
        (the ``*Buffer`` classes will insert inverters on :py:`o` and :py:`i` pins, as appropriate).

        This can be used for various purposes:

        - Normalizing active-low pins (such as ``CS_B``) to be active-high in Amaranth code
        - Compensating for boards where an inverting level-shifter (or similar circuitry) was used
          on the pin

        If the value is a simple :class:`bool`, it is used for all bits of this port. If the value
        is an iterable of :class:`bool`, the iterable must have the same length as :py:`io`, and
        the inversion is specified per-bit.
    direction : :class:`Direction` or :class:`str`
        Represents the allowed directions of this port. If equal to :attr:`Direction.Input` or
        :attr:`Direction.Output`, this port can only be used with buffers of matching direction.
        If equal to :attr:`Direction.Bidir`, this port can be used with buffers of any direction.
        If a string is passed, it is cast to :class:`Direction`.
    """
    def __init__(self, io: IOValueLike, *, invert : bool | Iterable[bool] = ..., direction: str | Direction = ...) -> None:
        ...

    @property
    def io(self) -> IOValue:
        """The :py:`io` argument passed to the constructor."""
        ...

    @property
    def invert(self) -> tuple[bool, ...]:
        """The :py:`invert` argument passed to the constructor, normalized to a :class:`tuple`
        of :class:`bool`."""
        ...

    @property
    def direction(self) -> Direction:
        """The :py:`direction` argument passed to the constructor, normalized to :class:`Direction`."""
        ...

    def __len__(self) -> int:
        """Returns the width of this port in bits. Equal to :py:`len(self.io)`."""
        ...

    def __invert__(self) -> SingleEndedPort:
        """Returns a new :class:`SingleEndedPort` with the opposite value of :py:`invert`."""
        ...

    def __getitem__(self, index: int | slice) -> SingleEndedPort:
        """Slices the port, returning another :class:`SingleEndedPort` with a subset
        of its bits.

        The index can be a :class:`slice` or :class:`int`. If the index is
        an :class:`int`, the result is a single-bit :class:`SingleEndedPort`."""
        ...

    def __add__(self, other: SingleEndedPort) -> SingleEndedPort:
        ...

    def __repr__(self) -> str:
        ...


class DifferentialPort(PortLike):
    """Represents a differential I/O port with optional inversion.

    Parameters
    ----------
    p : :class:`IOValue`
        The raw I/O value used as positive (true) half of the port.
    n : :class:`IOValue`
        The raw I/O value used as negative (complemented) half of the port. Must have the same
        length as :py:`p`.
    invert : :class:`bool` or iterable of :class`bool`
        If true, the electrical state of the physical pin will be opposite from the Amaranth value
        (the ``*Buffer`` classes will insert inverters on :py:`o` and :py:`i` pins, as appropriate).

        This can be used for various purposes:

        - Normalizing active-low pins (such as ``CS_B``) to be active-high in Amaranth code
        - Compensating for boards where the P and N pins are swapped (e.g. for easier routing)

        If the value is a simple :class:`bool`, it is used for all bits of this port. If the value
        is an iterable of :class:`bool`, the iterable must have the same length as :py:`io`, and
        the inversion is specified per-bit.
    direction : :class:`Direction` or :class:`str`
        Represents the allowed directions of this port. If equal to :attr:`Direction.Input` or
        :attr:`Direction.Output`, this port can only be used with buffers of matching direction.
        If equal to :attr:`Direction.Bidir`, this port can be used with buffers of any direction.
        If a string is passed, it is cast to :class:`Direction`.
    """
    def __init__(self, p: IOValueLike, n: IOValueLike, *, invert: bool | Iterable[bool] = ..., direction : str | Direction = ...) -> None:
        ...

    @property
    def p(self) -> IOValue:
        """The :py:`p` argument passed to the constructor."""
        ...

    @property
    def n(self) -> IOValue:
        """The :py:`n` argument passed to the constructor."""
        ...

    @property
    def invert(self) -> tuple[bool, ...]:
        """The :py:`invert` argument passed to the constructor, normalized to a :class:`tuple`
        of :class:`bool`."""
        ...

    @property
    def direction(self) -> Direction:
        """The :py:`direction` argument passed to the constructor, normalized to :class:`Direction`."""
        ...

    def __len__(self) -> int:
        """Returns the width of this port in bits. Equal to :py:`len(self.p)` (and :py:`len(self.n)`)."""
        ...

    def __invert__(self) -> DifferentialPort:
        """Returns a new :class:`DifferentialPort` with the opposite value of :py:`invert`."""
        ...

    def __getitem__(self, index: int | slice) -> DifferentialPort:
        """Slices the port, returning another :class:`DifferentialPort` with a subset
        of its bits.

        The index can be a :class:`slice` or :class:`int`. If the index is
        an :class:`int`, the result is a single-bit :class:`DifferentialPort`."""
        ...

    def __add__(self, other: DifferentialPort) -> DifferentialPort:
        ...

    def __repr__(self) -> str:
        ...


class SimulationPort(PortLike):
    """Represents a simulation library I/O port.
    Implements the :class:`PortLike` interface.
    Parameters
    ----------
    direction : :class:`Direction` or :class:`str`
        Set of allowed buffer directions. A string is converted to a :class:`Direction` first.
        If equal to :attr:`~Direction.Input` or :attr:`~Direction.Output`, this port can only be
        used with buffers of matching direction. If equal to :attr:`~Direction.Bidir`, this port
        can be used with buffers of any direction.
    width : :class:`int`
        Width of the port. The width of each of the attributes :py:`i`, :py:`o`, :py:`oe` (whenever
        present) equals :py:`width`.
    invert : :class:`bool` or iterable of :class:`bool`
        Polarity inversion. If the value is a simple :class:`bool`, it specifies inversion for
        the entire port. If the value is an iterable of :class:`bool`, the iterable must have the
        same length as the width of :py:`p` and :py:`n`, and the inversion is specified for
        individual wires.
    name : :class:`str` or :py:`None`
        Name of the port. This name is only used to derive the names of the input, output, and
        output enable signals.
    src_loc_at : :class:`int`
        :ref:`Source location <lang-srcloc>`. Used to infer :py:`name` if not specified.
    Attributes
    ----------
    i : :class:`Signal`
        Input signal. Present if :py:`direction in (Input, Bidir)`.
    o : :class:`Signal`
        Ouptut signal. Present if :py:`direction in (Output, Bidir)`.
    oe : :class:`Signal`
        Output enable signal. Present if :py:`direction in (Output, Bidir)`.
    invert : :class:`tuple` of :class:`bool`
        The :py:`invert` parameter, normalized to specify polarity inversion per-wire.
    direction : :class:`Direction`
        The :py:`direction` parameter, normalized to the :class:`Direction` enumeration.
    """
    def __init__(self, direction: str | Direction, width: int, *, invert: bool | Iterable[bool] = ..., name: Optional[str] = ..., src_loc_at: int = ...) -> None:
        ...

    @property
    def i(self) -> Signal:
        ...

    @property
    def o(self) -> Signal:
        ...

    @property
    def oe(self) -> Signal:
        ...

    @property
    def invert(self) -> tuple[bool, ...]:
        ...

    @property
    def direction(self) -> Direction:
        ...

    def __len__(self) -> int:
        ...

    def __getitem__(self, key: int | slice) -> SimulationPort:
        ...

    def __invert__(self) -> SimulationPort:
        ...

    def __add__(self, other) -> SimulationPort:
        ...

    def __repr__(self) -> str:
        ...


class Buffer(wiring.Component):
    """A combinatorial I/O buffer.

    Parameters
    ----------
    direction : :class:`Direction`
    port : :class:`PortLike`

    Attributes
    ----------
    signature : :class:`Buffer.Signature`
        Created based on constructor arguments.
    """
    class Signature(wiring.Signature):
        """A signature of a combinatorial I/O buffer.

        Parameters
        ----------
        direction : :class:`Direction`
        width : :class:`int`

        Attributes
        ----------
        i: :py:`unsigned(width)` (if :py:`direction in (Direction.Input, Direction.Bidir)`)
        o: :py:`unsigned(width)` (if :py:`direction in (Direction.Output, Direction.Bidir)`)
        oe: :py:`unsigned(1, init=0)` (if :py:`direction is Direction.Bidir`)
        oe: :py:`unsigned(1, init=1)` (if :py:`direction is Direction.Output`)
        """
        def __init__(self, direction: Direction, width: int) -> None:
            ...

        @property
        def direction(self) -> Direction:
            ...

        @property
        def width(self) -> int:
            ...

        def __eq__(self, other) -> bool:
            ...

        def __repr__(self) -> str:
            ...
    @property
    def i(self) -> Signal: ...

    @property
    def o(self) -> Signal: ...

    @property
    def oe(self) -> Signal: ...

    @property
    def signature(self) -> Signature:
        ...

    def __init__(self, direction: Direction, port: PortLike) -> None:
        ...

    @property
    def port(self) -> PortLike:
        ...

    @property
    def direction(self) -> Direction:
        ...

    def elaborate(self, platform) -> Module:
        ...


class FFBuffer(wiring.Component):
    """A registered I/O buffer.

    Equivalent to a plain :class:`Buffer` combined with reset-less registers on :py:`i`, :py:`o`,
    :py:`oe`.

    Parameters
    ----------
    direction : :class:`Direction`
    port : :class:`PortLike`
    i_domain : :class:`str`
        Domain for input register. Only used when :py:`direction in (Direction.Input, Direction.Bidir)`.
        Defaults to :py:`"sync"`
    o_domain : :class:`str`
        Domain for output and output enable registers. Only used when
        :py:`direction in (Direction.Output, Direction.Bidir)`. Defaults to :py:`"sync"`

    Attributes
    ----------
    signature : FFBuffer.Signature
        Created based on constructor arguments.
    """
    class Signature(wiring.Signature):
        """A signature of a registered I/O buffer.

        Parameters
        ----------
        direction : :class:`Direction`
        width : :class:`int`

        Attributes
        ----------
        i: :py:`unsigned(width)` (if :py:`direction in (Direction.Input, Direction.Bidir)`)
        o: :py:`unsigned(width)` (if :py:`direction in (Direction.Output, Direction.Bidir)`)
        oe: :py:`unsigned(1, init=0)` (if :py:`direction is Direction.Bidir`)
        oe: :py:`unsigned(1, init=1)` (if :py:`direction is Direction.Output`)
        """
        def __init__(self, direction: Direction, width: int) -> None:
            ...

        @property
        def direction(self) -> Direction:
            ...

        @property
        def width(self) -> int:
            ...

        def __eq__(self, other) -> bool:
            ...

        def __repr__(self) -> str:
            ...

    @property
    def signature(self) -> Signature:
        ...

    def __init__(self, direction: Direction, port: PortLike, *, i_domain: Optional[str] = ..., o_domain: Optional[str] = ...) -> None:
        if not isinstance(port, PortLike):
            raise TypeError(f"'port' must be a 'PortLike', not {port!r}")
        self._port = port
        super().__init__(FFBuffer.Signature(direction, len(port)).flip())
        if self.signature.direction is not Direction.Output:
            self._i_domain = i_domain or "sync"
        elif i_domain is not None:
            raise ValueError("Output buffer doesn't have an input domain")
        if self.signature.direction is not Direction.Input:
            self._o_domain = o_domain or "sync"
        elif o_domain is not None:
            raise ValueError("Input buffer doesn't have an output domain")
        if port.direction is Direction.Input and self.direction is not Direction.Input:
            raise ValueError(f"Input port cannot be used with {self.direction.name} buffer")
        if port.direction is Direction.Output and self.direction is not Direction.Output:
            raise ValueError(f"Output port cannot be used with {self.direction.name} buffer")

    @property
    def i(self) -> Signal: ...

    @property
    def o(self) -> Signal: ...

    @property
    def oe(self) -> Signal: ...

    @property
    def port(self) -> PortLike:
        ...

    @property
    def direction(self) -> Direction:
        ...

    @property
    def i_domain(self) -> str:
        ...

    @property
    def o_domain(self) -> str:
        ...

    def elaborate(self, platform) -> Module:
        ...


class DDRBuffer(wiring.Component):
    """A double data rate registered I/O buffer.

    In the input direction, the port is sampled at both edges of the input clock domain.
    The data sampled on the active clock edge of the domain appears on :py:`i[0]` with a delay
    of 1 clock cycle. The data sampled on the opposite clock edge appears on :py:`i[1]` with a delay
    of 0.5 clock cycle. Both :py:`i[0]` and :py:`i[1]` thus change on the active clock edge of the domain.

    In the output direction, both :py:`o[0]` and :py:`o[1]` are sampled on the active clock edge
    of the domain.  The value of :py:`o[0]` immediately appears on the output port.  The value
    of :py:`o[1]` then appears on the output port on the opposite edge, with a delay of 0.5 clock cycle.

    Support for this compoment is platform-specific, and may be missing on some platforms.

    Parameters
    ----------
    direction : :class:`Direction`
    port : :class:`PortLike`
    i_domain : :class:`str`
        Domain for input register. Only used when :py:`direction in (Direction.Input, Direction.Bidir)`.
    o_domain : :class:`str`
        Domain for output and output enable registers. Only used when
        :py:`direction in (Direction.Output, Direction.Bidir)`.

    Attributes
    ----------
    signature : DDRBuffer.Signature
        Created based on constructor arguments.
    """
    class Signature(wiring.Signature):
        """A signature of a double data rate registered I/O buffer.

        Parameters
        ----------
        direction : :class:`Direction`
        width : :class:`int`

        Attributes
        ----------
        i: :py:`unsigned(ArrayLayout(width, 2))` (if :py:`direction in (Direction.Input, Direction.Bidir)`)
        o: :py:`unsigned(ArrayLayout(width, 2))` (if :py:`direction in (Direction.Output, Direction.Bidir)`)
        oe: :py:`unsigned(1, init=0)` (if :py:`direction is Direction.Bidir`)
        oe: :py:`unsigned(1, init=1)` (if :py:`direction is Direction.Output`)
        """
        def __init__(self, direction: Direction, width: int) -> None:
            ...

        @property
        def direction(self) -> Direction:
            ...

        @property
        def width(self) -> int:
            ...

        def __eq__(self, other) -> bool:
            ...

        def __repr__(self) -> str:
            ...

    @property
    def signature(self) -> Signature:
        ...

    def __init__(self, direction: Direction, port: PortLike, *, i_domain: Optional[str] = ..., o_domain: Optional[str] = ...) -> None:
        ...

    @property
    def port(self) -> PortLike:
        ...

    @property
    def direction(self) -> Direction:
        ...

    @property
    def i_domain(self) -> str:
        ...

    @property
    def o_domain(self) -> str:
        ...

    def elaborate(self, platform):
        ...


class Pin(wiring.PureInterface):
    """
    An interface to an I/O buffer or a group of them that provides uniform access to input, output,
    or tristate buffers that may include a 1:n gearbox. (A 1:2 gearbox is typically called "DDR".)

    This is an interface object using :class:`Pin.Signature` as its signature.  The signature flows
    are defined from the point of view of a component that drives the I/O buffer.

    Parameters
    ----------
    width : int
        Width of the ``i``/``iN`` and ``o``/``oN`` signals.
    dir : ``"i"``, ``"o"``, ``"io"``, ``"oe"``
        Direction of the buffers. If ``"i"`` is specified, only the ``i``/``iN`` signals are
        present. If ``"o"`` is specified, only the ``o``/``oN`` signals are present. If ``"oe"`` is
        specified, the ``o``/``oN`` signals are present, and an ``oe`` signal is present.
        If ``"io"`` is specified, both the ``i``/``iN`` and ``o``/``oN`` signals are present, and
        an ``oe`` signal is present.
    xdr : int
        Gearbox ratio. If equal to 0, the I/O buffer is combinatorial, and only ``i``/``o``
        signals are present. If equal to 1, the I/O buffer is SDR, and only ``i``/``o`` signals are
        present. If greater than 1, the I/O buffer includes a gearbox, and ``iN``/``oN`` signals
        are present instead, where ``N in range(0, N)``. For example, if ``xdr=2``, the I/O buffer
        is DDR; the signal ``i0`` reflects the value at the rising edge, and the signal ``i1``
        reflects the value at the falling edge.
    path : tuple of str
        As in :class:`PureInterface`, used to name the created signals.

    Attributes
    ----------
    i_clk:
        I/O buffer input clock. Synchronizes `i*`. Present if ``xdr`` is nonzero.
    i_fclk:
        I/O buffer input fast clock. Synchronizes `i*` on higher gearbox ratios. Present if ``xdr``
        is greater than 2.
    i : Signal, out
        I/O buffer input, without gearing. Present if ``dir="i"`` or ``dir="io"``, and ``xdr`` is
        equal to 0 or 1.
    i0, i1, ... : Signal, out
        I/O buffer inputs, with gearing. Present if ``dir="i"`` or ``dir="io"``, and ``xdr`` is
        greater than 1.
    o_clk:
        I/O buffer output clock. Synchronizes `o*`, including `oe`. Present if ``xdr`` is nonzero.
    o_fclk:
        I/O buffer output fast clock. Synchronizes `o*` on higher gearbox ratios. Present if
        ``xdr`` is greater than 2.
    o : Signal, in
        I/O buffer output, without gearing. Present if ``dir="o"`` or ``dir="io"``, and ``xdr`` is
        equal to 0 or 1.
    o0, o1, ... : Signal, in
        I/O buffer outputs, with gearing. Present if ``dir="o"`` or ``dir="io"``, and ``xdr`` is
        greater than 1.
    oe : Signal, in
        I/O buffer output enable. Present if ``dir="io"`` or ``dir="oe"``. Buffers generally
        cannot change direction more than once per cycle, so at most one output enable signal
        is present.
    """

    @property
    def path(self) -> tuple[str | int, ...]:
        ...

    @property
    def name(self) -> str:
        ...

    @property
    def signature(self) -> Signature:
        ...

    class Signature(wiring.Signature):
        """A signature for :class:`Pin`.  The parameters are as defined on the ``Pin`` class,
        and are accessible as attributes.
        """
        def __init__(self, width: int, dir: str, *, xdr: int = ...) -> None:
            ...

        def __eq__(self, other) -> bool:
            ...

        def __repr__(self) -> str:
            ...

        def create(self, *, path: Optional[tuple[str | int, ...]] = ..., src_loc_at: int = ...) -> Pin:
            ...

    def __init__(self, width: int, dir: str, *, xdr: int = ..., name: Optional[str] = ..., path: Optional[tuple[str | int, ...]] = ..., src_loc_at: int = ...) -> None:
        ...

    @property
    def width(self) -> int:
        ...

    @property
    def dir(self) -> str:
        ...

    @property
    def xdr(self) -> int:
        ...



"""
This type stub file was generated by pyright.
"""

from ..hdl import *
from ..build import *

class LatticeICE40Platform(TemplatedPlatform):
    """
    .. rubric:: IceStorm toolchain

    Required tools:
        * ``yosys``
        * ``nextpnr-ice40``
        * ``icepack``

    The environment is populated by running the script specified in the environment variable
    ``AMARANTH_ENV_ICESTORM``, if present.

    Available overrides:
        * ``verbose``: enables logging of informational messages to standard error.
        * ``read_verilog_opts``: adds options for ``read_verilog`` Yosys command.
        * ``synth_opts``: adds options for ``synth_ice40`` Yosys command.
        * ``script_after_read``: inserts commands after ``read_ilang`` in Yosys script.
        * ``script_after_synth``: inserts commands after ``synth_ice40`` in Yosys script.
        * ``yosys_opts``: adds extra options for ``yosys``.
        * ``nextpnr_opts``: adds extra options for ``nextpnr-ice40``.
        * ``add_pre_pack``: inserts commands at the end in pre-pack Python script.
        * ``add_constraints``: inserts commands at the end in the PCF file.

    Build products:
        * ``{{name}}.rpt``: Yosys log.
        * ``{{name}}.json``: synthesized RTL.
        * ``{{name}}.tim``: nextpnr log.
        * ``{{name}}.asc``: ASCII bitstream.
        * ``{{name}}.bin``: binary bitstream.

    .. rubric:: iCECube2 toolchain

    This toolchain comes in two variants: ``LSE-iCECube2`` and ``Synplify-iCECube2``.

    Required tools:
        * iCECube2 toolchain
        * ``tclsh``

    The environment is populated by setting the necessary environment variables based on
    ``AMARANTH_ENV_ICECUBE2``, which must point to the root of the iCECube2 installation, and
    is required.

    Available overrides:
        * ``verbose``: enables logging of informational messages to standard error.
        * ``lse_opts``: adds options for LSE.
        * ``script_after_add``: inserts commands after ``add_file`` in Synplify Tcl script.
        * ``script_after_options``: inserts commands after ``set_option`` in Synplify Tcl script.
        * ``add_constraints``: inserts commands in SDC file.
        * ``script_after_flow``: inserts commands after ``run_sbt_backend_auto`` in SBT
          Tcl script.

    Build products:
        * ``{{name}}_lse.log`` (LSE) or ``{{name}}_design/{{name}}.htm`` (Synplify): synthesis log.
        * ``sbt/outputs/router/{{name}}_timing.rpt``: timing report.
        * ``{{name}}.edf``: EDIF netlist.
        * ``{{name}}.bin``: binary bitstream.
    """
    toolchain = ...
    device = ...
    package = ...
    _nextpnr_device_options = ...
    _nextpnr_package_options = ...
    _icestorm_required_tools = ...
    _icestorm_file_templates = ...
    _icestorm_command_templates = ...
    _icecube2_required_tools = ...
    _icecube2_file_templates = ...
    _lse_icecube2_command_templates = ...
    _synplify_icecube2_command_templates = ...
    def __init__(self, *, toolchain=...) -> None:
        ...
    
    @property
    def family(self): # -> Literal['iCE40', 'iCE5']:
        ...
    
    @property
    def required_tools(self): # -> list[str]:
        ...
    
    @property
    def file_templates(self): # -> dict[str, str]:
        ...
    
    @property
    def command_templates(self): # -> list[str]:
        ...
    
    @property
    def default_clk_constraint(self): # -> Clock:
        ...
    
    def create_missing_domain(self, name): # -> Module | None:
        ...
    
    def should_skip_port_component(self, port, attrs, component): # -> bool:
        ...
    
    def get_input(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_output(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_tristate(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_input_output(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_diff_input(self, pin, port, attrs, invert): # -> Module:
        ...
    
    def get_diff_output(self, pin, port, attrs, invert): # -> Module:
        ...
    





"""
This type stub file was generated by pyright.
"""

import enum
from typing import Any, Tuple
from .._utils import final
from .. import _unused
from . import _ast, _cd, _ir, _nir, _mem

"""
This type stub file was generated by pyright.
"""
__all__ = ["DomainRequirementFailed", "AlreadyElaborated", "UnusedElaboratable", "Elaboratable", "DuplicateElaboratable", "DriverConflict", "Fragment", "Instance", "IOBufferInstance", "RequirePosedge", "PortDirection", "Design", "build_netlist"]


@final
class AlreadyElaborated(Exception):
    """Exception raised when an elaboratable is being modified after elaboration."""


class UnusedElaboratable(_unused.UnusedMustUse):
    _MustUse__silence = ...


class Elaboratable(_unused.MustUse):
    ...


class DriverConflict(Exception):
    ...


class DuplicateElaboratable(Exception):
    ...


class DomainRequirementFailed(Exception):
    """Raised when a module has unsatisfied requirements about a clock domain, such as getting
    a negedge domain when only posedge domains are supported."""


class Fragment:
    @staticmethod
    def get(obj, platform):
        ...
    
    def __init__(self, *, src_loc=...) -> None:
        ...
    
    def add_domains(self, *domains):
        ...
    
    def iter_domains(self):
        ...
    
    def add_statements(self, domain, *stmts):
        ...
    
    def add_subfragment(self, subfragment, name=..., *, src_loc=...):
        ...
    
    def find_subfragment(self, name_or_index):
        ...
    
    def find_generated(self, *path):
        ...
    
    def elaborate(self, platform):
        ...
    
    def prepare(self, ports=..., *, hierarchy=..., missing_domain=..., propagate_domains=...):
        ...
    

class RequirePosedge(Fragment):
    """A special fragment that requires a given domain to have :py:`clk_edge="pos"`, failing
    elaboration otherwise.

    This is a private interface, without a stability guarantee.
    """
    def __init__(self, domain, *, src_loc_at=0, src_loc=None):
        ...


class Instance(Fragment):
    def __init__(self, type, *args, src_loc=..., src_loc_at=..., **kwargs) -> None:
        ...
    


class IOBufferInstance(Fragment):
    def __init__(self, port, *, i=..., o=..., oe=..., src_loc_at=..., src_loc=...) -> None:
        ...
    


class DesignFragmentInfo:
    def __init__(self, parent, depth) -> None:
        ...
    


class Design:
    """Represents a design ready for simulation or netlist building.

    Returned by ``Fragment.prepare``."""
    def __init__(self, fragment: Fragment, ports, *, hierarchy) -> None:
        ...
    


class PortDirection(enum.Enum):
    Input = ...
    Output = ...
    Inout = ...


class NetlistDriver:
    def __init__(self, module_idx: int, signal: _ast.Signal, domain: _cd.ClockDomain | None, *, src_loc) -> None:
        ...
    
    def emit_value(self, builder):
        ...
    


class NetlistEmitter:
    def __init__(self, netlist: _nir.Netlist, design, *, all_undef_to_ff=...) -> None:
        ...
    
    def emit_signal(self, signal) -> _nir.Value:
        ...
    
    def emit_io(self, value: _ast.IOValue) -> _nir.IOValue:
        ...
    
    def emit_io_use(self, value: _ast.IOValue, *, src_loc) -> _nir.IOValue:
        ...
    
    def emit_lhs(self, value: _ast.Value):
        ...
    
    def extend(self, value: _nir.Value, signed: bool, width: int):
        ...
    
    def emit_operator(self, module_idx: int, operator: str, *inputs: _nir.Value, src_loc):
        ...
    
    def unify_shapes_bitwise(self, operand_a: _nir.Value, signed_a: bool, operand_b: _nir.Value, signed_b: bool):
        ...
    
    def emit_rhs(self, module_idx: int, value: _ast.Value) -> Tuple[_nir.Value, bool]:
        """Emits a RHS value, returns a tuple of (value, is_signed)"""
        ...
    
    def connect(self, lhs: _nir.Value, rhs: _nir.Value, *, src_loc):
        ...
    
    def emit_assign(self, module_idx: int, cd: _cd.ClockDomain | None, lhs: _ast.Value, lhs_start: int, rhs: _nir.Value, cond: _nir.Net, *, src_loc):
        ...
    
    def emit_format(self, module_idx, format):
        ...
    
    def emit_stmt(self, module_idx: int, fragment: _ir.Fragment, domain: str, stmt: _ast.Statement, cond: _nir.Net):
        ...
    
    def emit_iobuffer(self, module_idx: int, instance: _ir.IOBufferInstance):
        ...
    
    def emit_memory(self, module_idx: int, fragment: _mem.MemoryInstance, name: str):
        ...
    
    def emit_write_port(self, module_idx: int, fragment: _mem.MemoryInstance, port: Any, memory: int):
        ...
    
    def emit_read_port(self, module_idx: int, fragment: _mem.MemoryInstance, port: Any, memory: int, write_ports: list[int]):
        ...
    
    def emit_instance(self, module_idx: int, instance: _ir.Instance, name: str):
        ...
    
    def emit_top_ports(self, fragment: _ir.Fragment):
        ...
    
    def emit_drivers(self):
        ...
    
    def emit_undef_ff(self):
        ...
    
    def emit_undriven(self):
        ...
    
    def emit_fragment(self, fragment: _ir.Fragment, parent_module_idx: int | None, *, cell_src_loc=...):
        ...
    


def build_netlist(fragment, ports=..., *, name=..., all_undef_to_ff=..., **kwargs):
    ...




"""
This type stub file was generated by pyright.
"""

import enum
from collections.abc import Iterable, Mapping, Iterator
from typing import NoReturn, Literal, TypeVar, Generic, Any, Self, Optional, overload
from ..hdl._ir import Elaboratable
from .._utils import final
from amaranth_types import ShapeLike, ValueLike, AbstractInterface, AbstractSignature, ModuleLike
from .meta import Annotation

__all__ = ["In", "Out", "Signature", "PureInterface", "connect", "flipped", "Component"]

_T_Signature = TypeVar("_T_Signature", bound=AbstractSignature)
_T_SignatureMembers = TypeVar("_T_SignatureMembers", bound=SignatureMembers)
_T_Interface = TypeVar("_T_Interface", bound=AbstractInterface)
_T = TypeVar("_T")

class Flow(enum.Enum):
    """Direction of data flow. This enumeration has two values, :attr:`Out` and :attr:`In`,
    the meaning of which depends on the context in which they are used.
    """
    Out = "out"
    In = "in"
    def flip(self) -> Flow:
        """Flip the direction of data flow.

        Returns
        -------
        :class:`Flow`
            :attr:`In` if called as :pc:`Out.flip()`; :attr:`Out` if called as :pc:`In.flip()`.
        """
        ...
    
    def __call__(self, description: AbstractSignature | ShapeLike, *, reset=...) -> Member:
        """Create a :class:`Member` with this data flow and the provided description and
        reset value.

        Returns
        -------
        :class:`Member`
            :pc:`Member(self, description, reset=reset)`
        """
        ...
    
    def __repr__(self) -> Literal['Out', 'In']:
        ...
    
    def __str__(self) -> str:
        ...
    


Out = Flow.Out
In = Flow.In

@final
class Member:
    """Description of a signature member.

    This class is a discriminated union: its instances describe either a `port member` or
    a `signature member`, and accessing properties for the wrong kind of member raises
    an :exc:`AttributeError`.

    The class is created from a `description`: a :class:`Signature` instance (in which case
    the :class:`Member` is created as a signature member), or a :ref:`shape-like <lang-shapelike>`
    object (in which case it is created as a port member). After creation the :class:`Member`
    instance cannot be modified.

    When a :class:`Signal` is created from a description of a port member, the signal's reset value
    is taken from the member description. If this signal is never explicitly assigned a value, it
    will equal ``reset``.

    Although instances can be created directly, most often they will be created through
    :data:`In` and :data:`Out`, e.g. :pc:`In(unsigned(1))` or :pc:`Out(stream.Signature(RGBPixel))`.
    """
    def __init__(self, flow: Flow, description: AbstractSignature | ShapeLike, *, reset=..., _dimensions=...) -> None:
        ...
    
    def flip(self) -> Member:
        """Flip the data flow of this member.

        Returns
        -------
        :class:`Member`
            A new :pc:`member` with :pc:`member.flow` equal to :pc:`self.flow.flip()`, and identical
            to :pc:`self` other than that.
        """
        ...
    
    def array(self, *dimensions) -> Member:
        """Add array dimensions to this member.

        The dimensions passed to this method are `prepended` to the existing dimensions.
        For example, :pc:`Out(1).array(2)` describes an array of 2 elements, whereas both
        :pc:`Out(1).array(2, 3)` and :pc:`Out(1).array(3).array(2)` both describe a two dimensional
        array of 2 by 3 elements.

        Dimensions are passed to :meth:`array` in the order in which they would be indexed.
        That is, :pc:`.array(x, y)` creates a member that can be indexed up to :pc:`[x-1][y-1]`.

        The :meth:`array` method is composable: calling :pc:`member.array(x)` describes an array of
        :pc:`x` members even if :pc:`member` was already an array.

        Returns
        -------
        :class:`Member`
            A new :pc:`member` with :pc:`member.dimensions` extended by :pc:`dimensions`, and
            identical to :pc:`self` other than that.
        """
        ...
    
    @property
    def flow(self) -> Flow:
        """Data flow of this member.

        Returns
        -------
        :class:`Flow`
        """
        ...
    
    @property
    def is_port(self) -> bool:
        """Whether this is a description of a port member.

        Returns
        -------
        :class:`bool`
            :pc:`True` if this is a description of a port member,
            :pc:`False` if this is a description of a signature member.
        """
        ...
    
    @property
    def is_signature(self) -> bool:
        """Whether this is a description of a signature member.

        Returns
        -------
        :class:`bool`
            :pc:`True` if this is a description of a signature member,
            :pc:`False` if this is a description of a port member.
        """
        ...
    
    @property
    def shape(self) -> ShapeLike:
        """Shape of a port member.

        Returns
        -------
        :ref:`shape-like object <lang-shapelike>`
            The shape that was provided when constructing this :class:`Member`.

        Raises
        ------
        :exc:`AttributeError`
            If :pc:`self` describes a signature member.
        """
        ...
    
    @property
    def reset(self): # -> None:
        """Reset value of a port member.

        Returns
        -------
        :ref:`const-castable object <lang-constcasting>`
            The reset value that was provided when constructing this :class:`Member`.

        Raises
        ------
        :exc:`AttributeError`
            If :pc:`self` describes a signature member.
        """
        ...
    
    @property
    def signature(self) -> AbstractSignature:
        """Signature of a signature member.

        Returns
        -------
        :class:`Signature`
            The signature that was provided when constructing this :class:`Member`.

        Raises
        ------
        :exc:`AttributeError`
            If :pc:`self` describes a port member.
        """
        ...
    
    @property
    def dimensions(self) -> tuple[int, ...]:
        """Array dimensions.

        A member will usually have no dimensions; in this case it does not describe an array.
        A single dimension describes one-dimensional array, and so on.

        Returns
        -------
        :class:`tuple` of :class:`int`
            Dimensions, if any, of this member, from most to least major.
        """
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def __repr__(self) -> str:
        ...
    


@final
class SignatureError(Exception):
    """
    This exception is raised when an invalid operation specific to signature manipulation is
    performed with :class:`SignatureMembers`, such as adding a member to a frozen signature.
    Other exceptions, such as :exc:`TypeError` or :exc:`NameError`, will still be raised where
    appropriate.
    """
    ...


@final
class SignatureMembers(Mapping[str, Member]):
    """Mapping of signature member names to their descriptions.

    This container, a :class:`collections.abc.Mapping`, is used to implement the :pc:`members`
    attribute of signature objects.

    The keys in this container must be valid Python attribute names that are public (do not begin
    with an underscore. The values must be instances of :class:`Member`. The container is mutable
    in a restricted manner: new keys may be added, but existing keys may not be modified or removed.
    In addition, the container can be `frozen`, which disallows addition of new keys. Freezing
    a container recursively freezes the members of any signatures inside.

    In addition to the use of the superscript operator, multiple members can be added at once with
    the :pc:`+=` opreator.

    The :meth:`create` method converts this mapping into a mapping of names to signature members
    (signals and interface objects) by creating them from their descriptions. The created mapping
    can be used to populate an interface object.
    """
    def __init__(self, members: Mapping[str, Member]=...) -> None:
        ...
    
    def flip(self) -> FlippedSignatureMembers[Self]:
        """Flip the data flow of the members in this mapping.

        Returns
        -------
        :class:`FlippedSignatureMembers`
            Proxy collection :pc:`FlippedSignatureMembers(self)` that flips the data flow of
            the members that are accessed using it.
        """
        ...
    
    def __eq__(self, other) -> bool:
        """Compare the members in this and another mapping.

        Returns
        -------
        :class:`bool`
            :pc:`True` if the mappings contain the same key-value pairs, :pc:`False` otherwise.
        """
        ...
    
    def __contains__(self, name: str) -> bool:
        """Check whether a member with a given name exists.

        Returns
        -------
        :class:`bool`
        """
        ...
    
    def __getitem__(self, name: str) -> Member:
        """Retrieves the description of a member with a given name.

        Returns
        -------
        :class:`Member`

        Raises
        ------
        :exc:`TypeError`
            If :pc:`name` is not a string.
        :exc:`NameError`
            If :pc:`name` is not a valid, public Python attribute name.
        :exc:`SignatureError`
            If a member called :pc:`name` does not exist in the collection.
        """
        ...
    
    def __setitem__(self, name: str, member: Member) -> NoReturn:
        """Stub that forbids addition of members to the collection.

        Raises
        ------
        :exc:`SignatureError`
            Always.
        """
        ...
    
    def __delitem__(self, name: str) -> NoReturn:
        """Stub that forbids removal of members from the collection.

        Raises
        ------
        :exc:`SignatureError`
            Always.
        """
        ...
    
    def __iter__(self) -> Iterator[str]:
        """Iterate through the names of members in the collection.

        Returns
        -------
        iterator of :class:`str`
            Names of members, in the order of insertion.
        """
        ...
    
    def __len__(self) -> int:
        ...
    
    def flatten(self, *, path: tuple[str | int, ...]=...) -> Iterator[tuple[tuple[str | int, ...], Member]]:
        """Recursively iterate through this collection.

        .. note::

            The :ref:`paths <wiring-path>` returned by this method and by :meth:`Signature.flatten`
            differ. This method yields a single result for each :class:`Member` in the collection,
            disregarding their dimensions:

            .. doctest::

                >>> sig = wiring.Signature({
                ...     "items": In(1).array(2)
                ... })
                >>> list(sig.members.flatten())
                [(('items',), In(1).array(2))]

            The :meth:`Signature.flatten` method yields multiple results for such a member; see
            the documentation for that method for an example.

        Returns
        -------
        iterator of (:class:`tuple` of :class:`str`, :class:`Member`)
            Pairs of :ref:`paths <wiring-path>` and the corresponding members. A path yielded by
            this method is a tuple of strings where each item is a key through which the item may
            be reached.
        """
        ...
    
    def create(self, *, path: tuple[str | int, ...] =..., src_loc_at: int =...) -> dict[str, Any]:
        """Create members from their descriptions.

        For each port member, this function creates a :class:`Signal` with the shape and reset
        value taken from the member description, and the name constructed from
        the :ref:`paths <wiring-path>` to the member (by concatenating path items with a double
        underscore, ``__``).

        For each signature member, this function calls :meth:`Signature.create` for that signature.
        The resulting object can have any type if a :class:`Signature` subclass overrides
        the :class:`create` method.

        If the member description includes dimensions, in each case, instead of a single member,
        a :class:`list` of members is created for each dimension. (That is, for a single dimension
        a list of members is returned, for two dimensions a list of lists is returned, and so on.)

        Returns
        -------
        dict of :class:`str` to :ref:`value-like <lang-valuelike>` or interface object or a potentially nested list of these
            Mapping of names to actual signature members.
        """
        ...
    
    def __repr__(self) -> str:
        ...
    


@final
class FlippedSignatureMembers(Mapping[str, Member], Generic[_T_SignatureMembers]):
    """Mapping of signature member names to their descriptions, with the directions flipped.

    Although an instance of :class:`FlippedSignatureMembers` could be created directly, it will
    be usually created by a call to :meth:`SignatureMembers.flip`.

    This container is a wrapper around :class:`SignatureMembers` that contains the same members
    as the inner mapping, but flips their data flow when they are accessed. For example:

    .. testcode::

        members = wiring.SignatureMembers({"foo": Out(1)})

        flipped_members = members.flip()
        assert flipped_members["foo"].flow == In

    This class implements the same methods, with the same functionality (other than the flipping of
    the data flow), as the :class:`SignatureMembers` class; see the documentation for that class
    for details.
    """
    def __init__(self, unflipped: _T_SignatureMembers) -> None:
        ...
    
    def flip(self) -> _T_SignatureMembers:
        """
        Flips this mapping back to the original one.

        Returns
        -------
        :class:`SignatureMembers`
            :pc:`unflipped`
        """
        ...
    
    def __eq__(self, other) -> bool:
        """Compare the members in this and another mapping.

        Returns
        -------
        :class:`bool`
            :pc:`True` if the mappings contain the same key-value pairs, :pc:`False` otherwise.
        """
        ...
    
    def __contains__(self, name: str) -> bool:
        ...
    
    def __getitem__(self, name: str) -> Member:
        ...
    
    def __setitem__(self, name: str, member: Member) -> NoReturn:
        ...
    
    def __delitem__(self, name: str) -> NoReturn:
        ...
    
    def __iter__(self) -> Iterator[str]:
        ...
    
    def __len__(self) -> int:
        ...
    
    def flatten(self, *, path: tuple[str | int, ...] = ...) -> Iterator[tuple[tuple[str | int, ...], Member]]:
        """Recursively iterate through this collection.

        .. note::

            The :ref:`paths <wiring-path>` returned by this method and by :meth:`Signature.flatten`
            differ. This method yields a single result for each :class:`Member` in the collection,
            disregarding their dimensions:

            .. doctest::

                >>> sig = wiring.Signature({
                ...     "items": In(1).array(2)
                ... })
                >>> list(sig.members.flatten())
                [(('items',), In(1).array(2))]

            The :meth:`Signature.flatten` method yields multiple results for such a member; see
            the documentation for that method for an example.

        Returns
        -------
        iterator of (:class:`tuple` of :class:`str`, :class:`Member`)
            Pairs of :ref:`paths <wiring-path>` and the corresponding members. A path yielded by
            this method is a tuple of strings where each item is a key through which the item may
            be reached.
        """
        ...
    
    def create(self, *, path: tuple[str | int, ...] =..., src_loc_at: int =...) -> dict[str, Any]:
        """Create members from their descriptions.

        For each port member, this function creates a :class:`Signal` with the shape and reset
        value taken from the member description, and the name constructed from
        the :ref:`paths <wiring-path>` to the member (by concatenating path items with a double
        underscore, ``__``).

        For each signature member, this function calls :meth:`Signature.create` for that signature.
        The resulting object can have any type if a :class:`Signature` subclass overrides
        the :class:`create` method.

        If the member description includes dimensions, in each case, instead of a single member,
        a :class:`list` of members is created for each dimension. (That is, for a single dimension
        a list of members is returned, for two dimensions a list of lists is returned, and so on.)

        Returns
        -------
        dict of :class:`str` to :ref:`value-like <lang-valuelike>` or interface object or a potentially nested list of these
            Mapping of names to actual signature members.
        """
        ...
    
    def __repr__(self) -> str:
        ...
    


class SignatureMeta(type):
    """Metaclass for :class:`Signature` that makes :class:`FlippedSignature` its
    'virtual subclass'.

    The object returned by :meth:`Signature.flip` is an instance of :class:`FlippedSignature`.
    It implements all of the methods :class:`Signature` has, and for subclasses of
    :class:`Signature`, it implements all of the methods defined on the subclass as well.
    This makes it effectively a subtype of :class:`Signature` (or a derived class of it), but this
    relationship is not captured by the Python type system: :class:`FlippedSignature` only has
    :class:`object` as its base class.

    This metaclass extends :func:`issubclass` and :func:`isinstance` so that they take into
    account the subtyping relationship between :class:`Signature` and :class:`FlippedSignature`,
    described below.
    """
    def __subclasscheck__(cls, subclass) -> bool:
        """
        Override of :pc:`issubclass(cls, Signature)`.

        In addition to the standard behavior of :func:`issubclass`, this override makes
        :class:`FlippedSignature` a subclass of :class:`Signature` or any of its subclasses.
        """
        ...
    
    def __instancecheck__(cls, instance) -> bool:
        """
        Override of :pc:`isinstance(obj, Signature)`.

        In addition to the standard behavior of :func:`isinstance`, this override makes
        :pc:`isinstance(obj, cls)` act as :pc:`isinstance(obj.flip(), cls)` where
        :pc:`obj` is an instance of :class:`FlippedSignature`.
        """
        ...
    


class Signature(metaclass=SignatureMeta):
    """Description of an interface object.

    An interface object is a Python object that has a :pc:`signature` attribute containing
    a :class:`Signature` object, as well as an attribute for every member of its signature.
    Signatures and interface objects are tightly linked: an interface object can be created out
    of a signature, and the signature is used when :func:`connect` ing two interface objects
    together. See the :ref:`introduction to interfaces <wiring-intro1>` for a more detailed
    explanation of why this is useful.

    :class:`Signature` can be used as a base class to define :ref:`customized <wiring-customizing>`
    signatures and interface objects.

    .. important::

        :class:`Signature` objects are immutable. Classes inheriting from :class:`Signature` must
        ensure this remains the case when additional functionality is added.
    """
    def __init__(self, members: Mapping[str, Member]) -> None:
        ...
    
    def flip(self) -> FlippedSignature[Self]:
        """Flip the data flow of the members in this signature.

        Returns
        -------
        :class:`FlippedSignature`
            Proxy object :pc:`FlippedSignature(self)` that flips the data flow of the attributes
            corresponding to the members that are accessed using it.

            See the documentation for the :class:`FlippedSignature` class for a detailed discussion
            of how this proxy object works.
        """
        ...
    
    @property
    def members(self) -> SignatureMembers:
        """Members in this signature.

        Returns
        -------
        :class:`SignatureMembers`
        """
        ...
    
    def __eq__(self, other) -> bool:
        """Compare this signature with another.

        The behavior of this operator depends on the types of the arguments. If both :pc:`self`
        and :pc:`other` are instances of the base :class:`Signature` class, they are compared
        structurally (the result is :pc:`self.members == other.members`); otherwise they are
        compared by identity (the result is :pc:`self is other`).

        Subclasses of :class:`Signature` are expected to override this method to take into account
        the specifics of the domain. If the subclass has additional properties that do not influence
        the :attr:`members` dictionary but nevertheless make its instance incompatible with other
        instances (for example, whether the feedback is combinational or registered),
        the overridden method must take that into account.

        Returns
        -------
        :class:`bool`
        """
        ...
    
    def flatten(self, obj) -> Iterator[tuple[tuple[str | int, ...], Flow, ValueLike]]:
        """Recursively iterate through this signature, retrieving member values from an interface
        object.

        .. note::

            The :ref:`paths <wiring-path>` returned by this method and by
            :meth:`SignatureMembers.flatten` differ. This method yield several results for each
            :class:`Member` in the collection that has a dimension:

            .. doctest::
                :options: +NORMALIZE_WHITESPACE

                >>> sig = wiring.Signature({
                ...     "items": In(1).array(2)
                ... })
                >>> obj = sig.create()
                >>> list(sig.flatten(obj))
                [(('items', 0), In(1), (sig obj__items__0)),
                 (('items', 1), In(1), (sig obj__items__1))]

            The :meth:`SignatureMembers.flatten` method yields one result for such a member; see
            the documentation for that method for an example.

        Returns
        -------
        iterator of (:class:`tuple` of :class:`str` or :class:`int`, :class:`Flow`, :ref:`value-like <lang-valuelike>`)
            Tuples of :ref:`paths <wiring-path>`, flow, and the corresponding member values. A path
            yielded by this method is a tuple of strings or integers where each item is an attribute
            name or index (correspondingly) using which the member value was retrieved.
        """
        ...
    
    def is_compliant(self, obj, *, reasons: Optional[list[str]] =..., path: tuple[str, ...] =...) -> bool:
        """Check whether an object matches the description in this signature.

        This module places few restrictions on what an interface object may be; it does not
        prescribe a specific base class or a specific way of constructing the object, only
        the values that its attributes should have. This method ensures consistency between
        the signature and the interface object, checking every aspect of the provided interface
        object for compliance with the signature.

        It verifies that:

        * :pc:`obj` has a :pc:`signature` attribute whose value a :class:`Signature` instance
          such that ``self == obj.signature``;
        * for each member, :pc:`obj` has an attribute with the same name, whose value:

          * for members with :meth:`dimensions <Member.dimensions>` specified, contains a list or
            a tuple (or several levels of nested lists or tuples, for multiple dimensions)
            satisfying the requirements below;
          * for port members, is a :ref:`value-like <lang-valuelike>` object casting to
            a :class:`Signal` or a :class:`Const` whose width and signedness is the same as that
            of the member, and (in case of a :class:`Signal`) which is not reset-less and whose
            reset value is that of the member;
          * for signature members, matches the description in the signature as verified by
            :meth:`Signature.is_compliant`.

        If the verification fails, this method reports the reason(s) by filling the :pc:`reasons`
        container. These reasons are intended to be human-readable: more than one reason may be
        reported but only in cases where this is helpful (e.g. the same error message will not
        repeat 10 times for each of the 10 ports in a list).

        Arguments
        ---------
        reasons : :class:`list` or :pc:`None`
            If provided, a container that receives diagnostic messages.
        path : :class:`tuple` of :class:`str`
            The :ref:`path <wiring-path>` to :pc:`obj`. Could be set to improve diagnostic
            messages if :pc:`obj` is nested within another object, or for clarity.

        Returns
        -------
        :class:`bool`
            :pc:`True` if :pc:`obj` matches the description in this signature, :pc:`False`
            otherwise. If :pc:`False` and :pc:`reasons` was not :pc:`None`, it will contain
            a detailed explanation why.
        """
        ...
    
    def create(self, *, path: tuple[str | int, ...]=..., src_loc_at: int =...) -> AbstractInterface[Self]:
        """Create an interface object from this signature.

        The default :meth:`Signature.create` implementation consists of one line:

        .. code::

            def create(self, *, path=None, src_loc_at=0):
                return PureInterface(self, path=path, src_loc_at=1 + src_loc_at)

        This implementation creates an interface object from this signature that serves purely
        as a container for the attributes corresponding to the signature members, and implements
        no behavior. Such an implementation is sufficient for signatures created ad-hoc using
        the :pc:`Signature({ ... })` constructor as well as simple signature subclasses.

        When defining a :class:`Signature` subclass that needs to customize the behavior of
        the created interface objects, override this method with a similar implementation
        that references the class of your custom interface object:

        .. testcode::

            class CustomSignature(wiring.Signature):
                def create(self, *, path=None, src_loc_at=0):
                    return CustomInterface(self, path=path, src_loc_at=1 + src_loc_at)

            class CustomInterface(wiring.PureInterface):
                @property
                def my_property(self):
                    ...

        The :pc:`path` and :pc:`src_loc_at` arguments are necessary to ensure the generated signals
        have informative names and accurate source location information.

        The custom :meth:`create` method may take positional or keyword arguments in addition to
        the two listed above. Such arguments must have a default value, because
        the :meth:`SignatureMembers.create` method will call the :meth:`Signature.create` member
        without these additional arguments when this signature is a member of another signature.
        """
        ...

    def annotations(self, obj: AbstractInterface[Self], /) -> Iterable[Annotation]:
        ...

    def __repr__(self) -> str:
        ...
    


@final
class FlippedSignature(Generic[_T_Signature]):
    """Description of an interface object, with the members' directions flipped.

    Although an instance of :class:`FlippedSignature` could be created directly, it will be usually
    created by a call to :meth:`Signature.flip`.

    This proxy is a wrapper around :class:`Signature` that contains the same description as
    the inner mapping, but flips the members' data flow when they are accessed. It is useful
    because :class:`Signature` objects are mutable and may include custom behavior, and if one was
    copied (rather than wrapped) by :meth:`Signature.flip`, the wrong object would be mutated, and
    custom behavior would be unavailable.

    For example:

    .. testcode::

        sig = wiring.Signature({"foo": Out(1)})

        flipped_sig = sig.flip()
        assert flipped_sig.members["foo"].flow == In

        sig.attr = 1
        assert flipped_sig.attr == 1
        flipped_sig.attr += 1
        assert sig.attr == flipped_sig.attr == 2

    This class implements the same methods, with the same functionality (other than the flipping of
    the members' data flow), as the :class:`Signature` class; see the documentation for that class
    for details.

    It is not possible to inherit from :class:`FlippedSignature` and :meth:`Signature.flip` must not
    be overridden. If a :class:`Signature` subclass defines a method and this method is called on
    a flipped instance of the subclass, it receives the flipped instance as its :pc:`self` argument.
    To distinguish being called on the flipped instance from being called on the unflipped one, use
    :pc:`isinstance(self, FlippedSignature)`:

    .. testcode::

        class SignatureKnowsWhenFlipped(wiring.Signature):
            @property
            def is_flipped(self):
                return isinstance(self, wiring.FlippedSignature)

        sig = SignatureKnowsWhenFlipped({})
        assert sig.is_flipped == False
        assert sig.flip().is_flipped == True
    """
    def __init__(self, signature: _T_Signature) -> None:
        ...
    
    def flip(self) -> _T_Signature:
        """
        Flips this signature back to the original one.

        Returns
        -------
        :class:`Signature`
            :pc:`unflipped`
        """
        ...
    
    @property
    def members(self) -> FlippedSignatureMembers:
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def flatten(self, obj) -> Iterator[tuple[tuple[str | int, ...], Flow, ValueLike]]:
        ...

    def is_compliant(self, obj, *, reasons: Optional[list[str]] =..., path: tuple[str, ...] =...) -> bool:
        ...

    def __getattr__(self, name) -> Any:
        """Retrieves attribute or method :pc:`name` of the unflipped signature.

        Performs :pc:`getattr(unflipped, name)`, ensuring that, if :pc:`name` refers to a property
        getter or a method, its :pc:`self` argument receives the *flipped* signature. A class
        method's :pc:`cls` argument receives the class of the *unflipped* signature, as usual.
        """
        ...
    
    def __setattr__(self, name, value) -> None:
        """Assigns attribute :pc:`name` of the unflipped signature to ``value``.

        Performs :pc:`setattr(unflipped, name, value)`, ensuring that, if :pc:`name` refers to
        a property setter, its :pc:`self` argument receives the flipped signature.
        """
        ...
    
    def __delattr__(self, name) -> None:
        """Removes attribute :pc:`name` of the unflipped signature.

        Performs :pc:`delattr(unflipped, name)`, ensuring that, if :pc:`name` refers to a property
        deleter, its :pc:`self` argument receives the flipped signature.
        """
        ...
    
    def create(self, *args, path: tuple[str | int, ...] =..., src_loc_at: int =..., **kwargs) -> FlippedInterface:
        ...
    
    def __repr__(self) -> str:
        ...
    


class PureInterface(Generic[_T_Signature]):
    """A helper for constructing ad-hoc interfaces.

    The :class:`PureInterface` helper primarily exists to be used by the default implementation of
    :meth:`Signature.create`, but it can also be used in any other context where an interface
    object needs to be created without the overhead of defining a class for it.

    .. important::

        Any object can be an interface object; it only needs a :pc:`signature` property containing
        a compliant signature. It is **not** necessary to use :class:`PureInterface` in order to
        create an interface object, but it may be used either directly or as a base class whenever
        it is convenient to do so.
    """
    signature: _T_Signature

    def __init__(self, signature: _T_Signature, *, path: tuple[str | int, ...]=..., src_loc_at: int =...) -> None:
        """Create attributes from a signature.

        The sole method defined by this helper is its constructor, which only defines
        the :pc:`self.signature` attribute as well as the attributes created from the signature
        members:

        .. code::

            def __init__(self, signature, *, path):
                self.__dict__.update({
                    "signature": signature,
                    **signature.members.create(path=path)
                })

        .. note::

            This implementation can be copied and reused in interface objects that *do* include
            custom behavior, if the signature serves as the source of truth for attributes
            corresponding to its members. Although it is less repetitive, this approach can confuse
            IDEs and type checkers.
        """
        ...
    
    def __repr__(self) -> str:
        ...
    


@final
class FlippedInterface(Generic[_T_Signature, _T_Interface]):
    """An interface object, with its members' directions flipped.

    An instance of :class:`FlippedInterface` should only be created by calling :func:`flipped`,
    which ensures that a :pc:`FlippedInterface(FlippedInterface(...))` object is never created.

    This proxy wraps any interface object and forwards attribute and method access to the wrapped
    interface object while flipping its signature and the values of any attributes corresponding to
    interface members. It is useful because interface objects may be mutable or include custom
    behavior, and explicitly keeping track of whether the interface object is flipped would be very
    burdensome.

    For example:

    .. testcode::

        intf = wiring.PureInterface(wiring.Signature({"foo": Out(1)}), path=())

        flipped_intf = wiring.flipped(intf)
        assert flipped_intf.signature.members["foo"].flow == In

        intf.attr = 1
        assert flipped_intf.attr == 1
        flipped_intf.attr += 1
        assert intf.attr == flipped_intf.attr == 2

    It is not possible to inherit from :class:`FlippedInterface`. If an interface object class
    defines a method or a property and it is called on the flipped interface object, the method
    receives the flipped interface object as its :pc:`self` argument. To distinguish being called
    on the flipped interface object from being called on the unflipped one, use
    :pc:`isinstance(self, FlippedInterface)`:

    .. testcode::

        class InterfaceKnowsWhenFlipped:
            signature = wiring.Signature({})

            @property
            def is_flipped(self):
                return isinstance(self, wiring.FlippedInterface)

        intf = InterfaceKnowsWhenFlipped()
        assert intf.is_flipped == False
        assert wiring.flipped(intf).is_flipped == True
    """
    def __init__(self, interface: _T_Interface) -> None:
        ...

    # not true -- this is a property -- but required for clean typing
    signature: _T_Signature
#    @property
#    def signature(self) -> _T_Signature:
#        """Signature of the flipped interface.
#
#        Returns
#        -------
#        Signature
#            :pc:`unflipped.signature.flip()`
#        """
#        ...
    
    def __eq__(self, other) -> bool:
        """Compare this flipped interface with another.

        Returns
        -------
        bool
            :pc:`True` if :pc:`other` is an instance :pc:`FlippedInterface(other_unflipped)` where
            :pc:`unflipped == other_unflipped`, :pc:`False` otherwise.
        """
        ...
    
    def __getattr__(self, name) -> Any:
        """Retrieves attribute or method :pc:`name` of the unflipped interface.

        Performs :pc:`getattr(unflipped, name)`, with the following caveats:

        1. If :pc:`name` refers to a signature member, the returned interface object is flipped.
        2. If :pc:`name` refers to a property getter or a method, its :pc:`self` argument receives
           the *flipped* interface. A class method's :pc:`cls` argument receives the class of
           the *unflipped* interface, as usual.
        """
        ...
    
    def __setattr__(self, name, value) -> None:
        """Assigns attribute :pc:`name` of the unflipped interface to ``value``.

        Performs :pc:`setattr(unflipped, name, value)`, with the following caveats:

        1. If :pc:`name` refers to a signature member, the assigned interface object is flipped.
        2. If :pc:`name` refers to a property setter, its :pc:`self` argument receives the flipped
           interface.
        """
        ...
    
    def __delattr__(self, name) -> None:
        """Removes attribute :pc:`name` of the unflipped interface.

        Performs :pc:`delattr(unflipped, name)`, ensuring that, if :pc:`name` refers to a property
        deleter, its :pc:`self` argument receives the flipped interface.
        """
        ...
    
    def __repr__(self) -> str:
        ...
    
@overload
def flipped(interface: FlippedInterface[_T_Signature, _T_Interface]) -> _T_Interface:
    ...

# Can't be typed nicer for now.
@overload
def flipped(interface: _T_Interface) -> FlippedInterface[Any, _T_Interface]:
    ...

def flipped(interface: _T_Interface) -> _T_Interface | FlippedInterface[Any, _T_Interface]:
    """
    Flip the data flow of the members of the interface object :pc:`interface`.

    If an interface object is flipped twice, returns the original object:
    :pc:`flipped(flipped(interface)) is interface`. Otherwise, wraps :pc:`interface` in
    a :class:`FlippedInterface` proxy object that flips the directions of its members.

    See the documentation for the :class:`FlippedInterface` class for a detailed discussion of how
    this proxy object works.
    """
    ...

@final
class ConnectionError(Exception):
    """Exception raised when the :func:`connect` function is requested to perform an impossible,
    meaningless, or forbidden connection."""
    ...


def connect(m: ModuleLike, *args: AbstractInterface, **kwargs: AbstractInterface) -> None:
    """Connect interface objects to each other.

    This function creates connections between ports of several interface objects. (Any number of
    interface objects may be provided; in most cases it is two.)

    The connections can be made only if all of the objects satisfy a number of requirements:

    * Every interface object must have the same set of port members, and they must have the same
      :meth:`dimensions <Member.dimensions>`.
    * For each path, the port members of every interface object must have the same width and reset
      value (for port members corresponding to signals) or constant value (for port members
      corresponding to constants). Signedness may differ.
    * For each path, at most one interface object must have the corresponding port member be
      an output.
    * For a given path, if any of the interface objects has an input port member corresponding
      to a constant value, then the rest of the interface objects must have output port members
      corresponding to the same constant value.

    For example, if :pc:`obj1` is being connected to :pc:`obj2` and :pc:`obj3`, and :pc:`obj1.a.b`
    is an output, then :pc:`obj2.a.b` and :pc:`obj2.a.b` must exist and be inputs. If :pc:`obj2.c`
    is an input and its value is :pc:`Const(1)`, then :pc:`obj1.c` and :pc:`obj3.c` must be outputs
    whose value is also :pc:`Const(1)`. If no ports besides :pc:`obj1.a.b` and :pc:`obj1.c` exist,
    then no ports except for those two must exist on :pc:`obj2` and :pc:`obj3` either.

    Once it is determined that the interface objects can be connected, this function performs
    an equivalent of:

    .. code::

        m.d.comb += [
            in1.eq(out1),
            in2.eq(out1),
            ...
        ]

    Where :pc:`out1` is an output and :pc:`in1`, :pc:`in2`, ... are the inputs that have the same
    path. (If no interface object has an output for a given path, **no connection at all** is made.)

    The positions (within :pc:`args`) or names (within :pc:`kwargs`) of the arguments do not affect
    the connections that are made. There is no difference in behavior between :pc:`connect(m, a, b)`
    and :pc:`connect(m, b, a)` or :pc:`connect(m, arbiter=a, decoder=b)`. The names of the keyword
    arguments serve only a documentation purpose: they clarify the diagnostic messages when
    a connection cannot be made.
    """
    ...

class Component(Elaboratable, Generic[_T_Signature]):
    """Base class for elaboratable interface objects.

    A component is an :class:`Elaboratable` whose interaction with other parts of the design is
    defined by its signature. Most if not all elaboratables in idiomatic Amaranth code should be
    components, as the signature clarifies the direction of data flow at their boundary. See
    the :ref:`introduction to interfaces <wiring-intro1>` section for a practical guide to defining
    and using components.

    There are two ways to define a component. If all instances of a component have the same
    signature, it can be defined using :term:`variable annotations <python:variable annotation>`:

    .. testcode::

        class FixedComponent(wiring.Component):
            en: In(1)
            data: Out(8)

    The variable annotations are collected by the constructor :meth:`Component.__init__`. Only
    public (not starting with ``_``) annotations with :class:`In <Member>` or :class:`Out <Member>`
    objects are considered; all other annotations are ignored under the assumption that they are
    interpreted by some other tool.

    It is possible to use inheritance to extend a component: the component's signature is composed
    from the variable annotations in the class that is being constructed as well as all of its
    base classes. It is an error to have more than one variable annotation for the same attribute.

    If different instances of a component may need to have different signatures, variable
    annotations cannot be used. In this case, the constructor should be overridden, and
    the computed signature members should be provided to the superclass constructor:

    .. testcode::

        class ParametricComponent(wiring.Component):
            def __init__(self, data_width):
                super().__init__({
                    "en": In(1),
                    "data": Out(data_width)
                })

    It is also possible to pass a :class:`Signature` instance to the superclass constructor.

    Aside from initializing the :attr:`signature` attribute, the :meth:`Component.__init__`
    constructor creates attributes corresponding to all of the members defined in the signature.
    If an attribute with the same name as that of a member already exists, an error is raied.

    Raises
    ------
    :exc:`TypeError`
        If the :pc:`signature` object is neither a :class:`Signature` nor a :class:`dict`.
        If neither variable annotations nor the :pc:`signature` argument are present, or if
        both are present.
    :exc:`NameError`
        If a name conflict is detected between two variable annotations, or between a member
        and an existing attribute.
    """
    def __init__(self, signature: Optional[_T_Signature | dict[str, Member]] = None) -> None:
        ...
    
    @property
    def signature(self) -> _T_Signature:
        """The signature of the component.

        .. important::

            Do not override this property. Once a component is constructed, its :attr:`signature`
            property must always return the same :class:`Signature` instance. The constructor
            can be used to customize a component's signature.
        """
        ...

    @property
    def metadata(self) -> ComponentMetadata:
        """Metadata attached to the component.
        
        Returns
        -------
        :class:`ComponentMetadata`
        """
        ...


class InvalidMetadata(Exception):
    """Exception raised by :meth:`ComponentMetadata.validate` when the JSON representation of
    a component's metadata does not conform to its schema."""


class ComponentMetadata(Annotation[Component]):
    def __init__(self, origin: Component):
        ...

    @property
    def origin(self) -> Component:
        ...

    @classmethod
    def validate(cls, instance: dict) -> None:
        ...

    def as_json(self) -> dict:
        ...



"""
This type stub file was generated by pyright.
"""

from typing import Any, Optional
from ..hdl import *
from .._utils import final
from . import wiring

@final
class Signature(wiring.Signature):
    """Signature of a unidirectional data stream.

    .. note::

        "Minimal streams" as defined in `RFC 61`_ lack support for complex payloads, such as
        multiple lanes or packetization, as well as introspection of the payload. This limitation
        will be lifted in a later release.

        .. _RFC 61: https://amaranth-lang.org/rfcs/0061-minimal-streams.html

    Parameters
    ----------
    payload_shape : :class:`~.hdl.ShapeLike`
        Shape of the payload.
    always_valid : :class:`bool`
        Whether the stream has a payload available each cycle.
    always_ready : :class:`bool`
        Whether the stream has its payload accepted whenever it is available (i.e. whether it lacks
        support for backpressure).

    Members
    -------
    payload : :py:`Out(payload_shape)`
        Payload.
    valid : :py:`Out(1)`
        Whether a payload is available. If the stream is :py:`always_valid`, :py:`Const(1)`.
    ready : :py:`In(1)`
        Whether a payload is accepted. If the stream is :py:`always_ready`, :py:`Const(1)`.
    """
    def __init__(self, payload_shape: ShapeLike, *, payload_init: Any = ..., always_valid: bool=..., always_ready: bool=...) -> None:
        ...
    
    @property
    def always_valid(self) -> bool:
        ...
    
    @property
    def always_ready(self) -> bool:
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def create(self, *, path: Optional[tuple[str | int, ...]] = ..., src_loc_at: int = ...) -> Interface:
        ...
    
    def __repr__(self) -> str:
        ...
    


@final
class Interface:
    """A unidirectional data stream.

    Attributes
    ----------
    signature : :class:`Signature`
        Signature of this data stream.
    """
    payload: Any
    valid: Value
    ready: Value
    def __init__(self, signature: Signature, *, path: Optional[tuple[str | int, ...]] = ..., src_loc_at: int =...) -> None:
        ...
    
    @property
    def signature(self) -> Signature:
        ...
    
    @property
    def p(self) -> Any:
        """Shortcut for :py:`self.payload`.

        This shortcut reduces repetition when manipulating the payload, for example:

        .. code::

            m.d.comb += [
                self.o_stream.p.result.eq(self.i_stream.p.first + self.i_stream.p.second),
                self.o_stream.valid.eq(self.i_stream.valid),
                self.i_stream.ready.eq(self.o_stream.ready),
            ]
        """
        ...
    
    def __repr__(self) -> str:
        ...
    





"""
This type stub file was generated by pyright.
"""

import enum as py_enum
from typing import Generic, TypeVar, Self, overload
from amaranth import *
from ..hdl._ast import Assign, ValueCastable, ShapeCastable, ShapeLike, ValueLike, Format

__all__ = ['EnumType', 'Enum', 'IntEnum', 'Flag', 'IntFlag', 'EnumView', 'FlagView', 'auto', 'unique']


_T = TypeVar("_T")
_T_ViewClass = TypeVar("_T_ViewClass", bound=None | ValueCastable)


auto = py_enum.auto
unique = py_enum.unique


class EnumType(ShapeCastable, py_enum.EnumType, Generic[_T_ViewClass]):
    """Subclass of the standard :class:`enum.EnumType` that implements the :class:`ShapeCastable`
    protocol.

    This metaclass provides the :meth:`as_shape` method, making its instances
    :ref:`shape-castable <lang-shapecasting>`, and accepts a ``shape=`` keyword argument
    to specify a shape explicitly. Other than this, it acts the same as the standard
    :class:`enum.EnumType` class; if the ``shape=`` argument is not specified and
    :meth:`as_shape` is never called, it places no restrictions on the enumeration class
    or the values of its members.
    """
    @classmethod
    def __prepare__(metacls, name, bases, shape: ShapeLike =..., view_class:_T_ViewClass=..., **kwargs) -> py_enum._EnumDict:
        ...
    
    def __new__(cls, name, bases, namespace, shape: ShapeLike =..., view_class:_T_ViewClass=..., **kwargs) -> Self:
        ...
    
    def as_shape(cls) -> Shape:
        """Cast this enumeration to a shape.

        Returns
        -------
        :class:`Shape`
            Explicitly provided shape. If not provided, returns the result of shape-casting
            this class :ref:`as a standard Python enumeration <lang-shapeenum>`.

        Raises
        ------
        TypeError
            If the enumeration has neither an explicitly provided shape nor any members.
        """
        ...
    
    @overload
    def __call__(cls: type[_T], value: int) -> _T:
        ...

    @overload
    def __call__(cls: type[_T], value: _T) -> _T:
        ...

    @overload
    def __call__(cls: EnumType[None], value: int | ValueLike) -> Value:
        ...

    @overload
    def __call__(cls: EnumType[_T_ViewClass], value: int | ValueLike) -> _T_ViewClass:
        ...

    def __call__(cls, value: int | ValueLike) -> Value | ValueCastable:
        ...
    
    def const(cls: EnumType[_T_ViewClass], init) -> _T_ViewClass:
        ...

    def from_bits(cls: EnumType[_T_ViewClass], bits: int) -> _T_ViewClass:
        ...

    def format(cls, value: ValueLike, format_spec: str) -> Format.Enum:
        ...
    

class Enum(py_enum.Enum, metaclass=EnumType[EnumView]):
    """Subclass of the standard :class:`enum.Enum` that has :class:`EnumType` as
    its metaclass."""
    ...


class IntEnum(py_enum.IntEnum, metaclass=EnumType[None]):
    """Subclass of the standard :class:`enum.IntEnum` that has :class:`EnumType` as
    its metaclass."""
    ...


class Flag(py_enum.Flag, metaclass=EnumType[FlagView]):
    """Subclass of the standard :class:`enum.Flag` that has :class:`EnumType` as
    its metaclass."""
    ...


class IntFlag(py_enum.IntFlag, metaclass=EnumType[None]):
    """Subclass of the standard :class:`enum.IntFlag` that has :class:`EnumType` as
    its metaclass."""
    ...


class EnumView(ValueCastable, Generic[_T]):
    """The view class used for :class:`Enum`.

    Wraps a :class:`Value` and only allows type-safe operations. The only operators allowed are
    equality comparisons (``==`` and ``!=``) with another :class:`EnumView` of the same enum type.
    """

    def __init__(self, enum: _T, target: ValueLike):
        ...

    def shape(self) -> _T:
        ...

    @ValueCastable.lowermethod
    def as_value(self) -> Value:
        ...

    def eq(self, other: ValueLike) -> Assign:
        ...

    def __eq__(self, other: FlagView[_T] | _T) -> Value:
        """Compares the underlying value for equality.

        The other operand has to be either another :class:`EnumView` with the same enum type, or
        a plain value of the underlying enum.

        Returns
        -------
        :class:`Value`
            The result of the equality comparison, as a single-bit value.
        """
        ...

    def __ne__(self, other: FlagView[_T] | _T) -> Value:
        ...



class FlagView(EnumView[_T], Generic[_T]):
    """The view class used for :class:`Flag`.

    In addition to the operations allowed by :class:`EnumView`, it allows bitwise operations among
    values of the same enum type."""

    def __invert__(self) -> FlagView[_T]:
        """Inverts all flags in this value and returns another :ref:`FlagView`.

        Note that this is not equivalent to applying bitwise negation to the underlying value:
        just like the Python :class:`enum.Flag` class, only bits corresponding to flags actually
        defined in the enumeration are included in the result.

        Returns
        -------
        :class:`FlagView`
        """
        ...

    def __and__(self, other: FlagView[_T] | _T) -> FlagView[_T]:
        """Performs a bitwise AND and returns another :class:`FlagView`.

        The other operand has to be either another :class:`FlagView` of the same enum type, or
        a plain value of the underlying enum type.

        Returns
        -------
        :class:`FlagView`
        """
        ...

    def __or__(self, other: FlagView[_T] | _T) -> FlagView[_T]:
        """Performs a bitwise OR and returns another :class:`FlagView`.

        The other operand has to be either another :class:`FlagView` of the same enum type, or
        a plain value of the underlying enum type.

        Returns
        -------
        :class:`FlagView`
        """
        ...

    def __xor__(self, other: FlagView[_T] | _T) -> FlagView[_T]:
        """Performs a bitwise XOR and returns another :class:`FlagView`.

        The other operand has to be either another :class:`FlagView` of the same enum type, or
        a plain value of the underlying enum type.

        Returns
        -------
        :class:`FlagView`
        """
        ...

    __rand__ = __and__
    __ror__ = __or__
    __rxor__ = __xor__






"""
This type stub file was generated by pyright.
"""

from enum import Enum
from typing import NoReturn, OrderedDict
from ._ast import *
from amaranth_types import LayoutLike

__all__ = ["Direction", "DIR_NONE", "DIR_FANOUT", "DIR_FANIN", "Layout", "Record"]
Direction = Enum('Direction', ('NONE', 'FANOUT', 'FANIN'))
DIR_NONE = Direction.NONE
DIR_FANOUT = Direction.FANOUT
DIR_FANIN = Direction.FANIN
class Layout:
    fields: OrderedDict

    @staticmethod
    def cast(obj: LayoutLike, *, src_loc_at=...) -> Layout:
        ...
    
    def __init__(self, fields, *, src_loc_at=...) -> None:
        ...
    
    def __getitem__(self, item): # -> Layout:
        ...
    
    def __iter__(self): # -> Generator[tuple[Unknown, Unknown, Unknown], None, None]:
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def __repr__(self) -> str:
        ...
    


class Record(ValueCastable):
    name: str
    layout: Layout
    fields: OrderedDict

    @staticmethod
    def like(other: Record, *, name=..., name_suffix=..., src_loc_at=...) -> Record:
        ...
    
    def __init__(self, layout: LayoutLike, *, name=..., fields=..., src_loc_at=...) -> None:
        ...

    def __bool__(self) -> NoReturn:
        ...

    def __invert__(self) -> Value:
        ...

    def __neg__(self) -> Value:
        ...

    def __add__(self, other: ValueLike) -> Value:
        ...

    def __radd__(self, other: ValueLike) -> Value:
        ...

    def __sub__(self, other: ValueLike) -> Value:
        ...

    def __rsub__(self, other: ValueLike) -> Value:
        ...

    def __mul__(self, other: ValueLike) -> Value:
        ...

    def __rmul__(self, other: ValueLike) -> Value:
        ...

    def __mod__(self, other: ValueLike) -> Value:
        ...

    def __rmod__(self, other: ValueLike) -> Value:
        ...

    def __floordiv__(self, other: ValueLike) -> Value:
        ...

    def __rfloordiv__(self, other: ValueLike) -> Value:
        ...

    def __lshift__(self, other: ValueLike) -> Value:
        ...

    def __rlshift__(self, other: ValueLike) -> Value:
        ...

    def __rshift__(self, other: ValueLike) -> Value:
        ...

    def __rrshift__(self, other: ValueLike) -> Value:
        ...

    def __and__(self, other: ValueLike) -> Value:
        ...

    def __rand__(self, other: ValueLike) -> Value:
        ...

    def __xor__(self, other: ValueLike) -> Value:
        ...

    def __rxor__(self, other: ValueLike) -> Value:
        ...

    def __or__(self, other: ValueLike) -> Value:
        ...

    def __ror__(self, other: ValueLike) -> Value:
        ...

    def __eq__(self, other: ValueLike) -> Value:
        ...

    def __ne__(self, other: ValueLike) -> Value:
        ...

    def __lt__(self, other: ValueLike) -> Value:
        ...

    def __le__(self, other: ValueLike) -> Value:
        ...

    def __gt__(self, other: ValueLike) -> Value:
        ...

    def __ge__(self, other: ValueLike) -> Value:
        ...

    def __abs__(self) -> Value:
        ...

    def as_unsigned(self) -> Value:
        ...

    def as_signed(self) -> Value:
        ...
    
    def bool(self) -> Value:
        ...

    def any(self) -> Value:
        ...

    def all(self) -> Value:
        ...

    def xor(self) -> Value:
        ...

    def implies(premise, conclusion: ValueLike) -> Value:
        ...

    def bit_select(self, offset: ValueLike, width: int) -> Value:
        ...

    def word_select(self, offset: ValueLike, width: int) -> Value:
        ...

    def matches(self, *patterns) -> Value:
        ...

    def shift_left(self, amount: int) -> Value:
        ...

    def shift_right(self, amount: int) -> Value:
        ...

    def rotate_left(self, amount: int) -> Value:
        ...

    def rotate_right(self, amount: int) -> Value:
        ...

    def eq(self, value: ValueLike) -> Assign:
        ...

    def __getattr__(self, name: str):
        ...
    
    def __getitem__(self, item: str | tuple | int | slice):
        ...
    
    @ValueCastable.lowermethod
    def as_value(self) -> Value:
        ...
    
    def __len__(self) -> int:
        ...
    
    def __repr__(self) -> str:
        ...
    
    def shape(self) -> Shape:
        ...
    
    def connect(self, *subordinates, include=..., exclude=...):
        ...
    





"""
This type stub file was generated by pyright.
"""

__all__ = ["flatten", "union", "memoize", "final", "deprecated", "get_linter_options", "get_linter_option"]

def flatten(i): # -> Generator[str | Unknown, Unknown, None]:
    ...

def union(i, start=...): # -> None:
    ...

def memoize(f): # -> _Wrapped[..., Unknown, (*args: Unknown), Unknown]:
    ...

def final(cls):
    ...

def deprecated(message, stacklevel=...): # -> (f: Unknown) -> _Wrapped[..., Unknown, (*args: Unknown, **kwargs: Unknown), Unknown]:
    ...

def get_linter_options(filename): # -> dict[str, str] | dict[Unknown, Unknown]:
    ...

def get_linter_option(filename, name, type, default): # -> int | bool:
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Callable

from .._utils import deprecated
from ..hdl._cd import *
from ..hdl._ir import *
from ._pycoro import Tick, Settle, Delay, Passive, Active
from ._async import DomainReset, BrokenTrigger, ProcessContext, TestbenchContext
from amaranth_types import TestGenerator, TestCoroutine

__all__ = [
    "DomainReset", "BrokenTrigger",
    "Simulator",
    # deprecated
    "Settle", "Delay", "Tick", "Passive", "Active",
]

class Simulator:
    def __init__(self, fragment, *, engine=...) -> None:
        ...
    
    def add_process(self, process: Callable[[], TestGenerator[None]] | Callable[[ProcessContext], TestCoroutine[None]]) -> None:
        ...
    
    def add_testbench(self, process: Callable[[], TestGenerator[None]] | Callable[[TestbenchContext], TestCoroutine[None]], *, background: bool = False) -> None:
        ...

    def add_clock(self, period, *, phase=..., domain=..., if_exists=...):
        """Add a clock process.

        Ad"""
        ...
    
    def reset(self): # -> None:
        """Reset the simulation.

        A"""
        ...
    
    @deprecated("instead of `sim.step()`, use `si")
    def step(self): # -> bool:
        ...
    
    def advance(self): # -> bool:
        """Advance the simulation.

       """
        ...
    
    def run(self): # -> None:
        """Run the simulation while any pro"""
        ...
    
    def run_until(self, deadline, *, run_passive=...): # -> None:
        """Run the simulation until it adva"""
        ...
    
    def write_vcd(self, vcd_file, gtkw_file=..., *, traces=..., fs_per_delta=...): # -> _GeneratorContextManager[None]:
        """Write waveforms to a Value Chang"""
        ...
    


